'''For dynamically inspecting and modifying attributes of Python objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Optional, Union
import re


def compile_argfree_getable_attrs(obj : Any, getter_re : Union[str, re.Pattern]='.*', repl_str : Optional[str]=None) -> dict[str, Any]:
    '''Compile the values of all methods of an object which require no arguments other than perhaps the object itself (this EXCLUDES properties)
    Returns a dict whose keys are the names of the methods called and whose values are the return values of those object methods

    Can optionally filter the names of returned method using a regular expression, passed to "getter_re"
    Can also optionally replace the chosen regex with an arbitrary string (including the empty string), passed to "repl_str"
    
    Parameters
    ----------
    obj : Any
        Any object instance
    getter_re : str or re.Pattern (optional), default ".*"
        Optional regular expression to use for filtering down returned methods
        Only methods whose names match the target regex are returns
    repl_str : str (optional)
        If provided, will replace the 
        for example, repl_str="" can be used to delete the regex from returned method names

    Returns
    -------
    getable_dict : dict[str, Any]
        dict whose keys are the selected method names and whose values are the corresponding method returns
    '''
    getable_dict = {}
    for attr_name in dir(obj):
        if re.search(getter_re, attr_name): 
            try:
                attr_key = attr_name if (repl_str is None) else re.sub(getter_re, repl_str, attr_name)
                getable_dict[attr_key] = getattr(obj, attr_name)()
            except (TypeError, Exception): # TODO : find way to selectively intercept the Boost C++ wrapper ArgumentError
                pass
    return getable_dict


'''Tabulated references of Greek letter names, prefices, and unicode characters'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

GREEK_LETTER_NAMES = [ # names for greek character literals
    'alpha',
    'beta',
    'gamma',
    'delta',
    'epsilon',
    'zeta',
    'eta',
    'theta',
    'iota',
    'kappa',
    'lambda',
    'mu',
    'nu',
    'xi',
    'omicron',
    'pi',
    'rho',
    'sigma_end',
    'sigma',
    'tau',
    'upsilon',
    'phi',
    'chi',
    'psi',
    'omega'
]

_GREEK_START_IDXS = { # indices where each case of the Greek alphabet starts in Unicode
    'LOWER' : 945,
    'UPPER' : 913
}

for case, idx in _GREEK_START_IDXS.items():
    globals()[f'GREEK_{case}'] = { # add dicts to global namespace
        letter_name : chr(idx + i)
            for i, letter_name in enumerate(GREEK_LETTER_NAMES)
    }


'''Utilities for examining prime numbers and integer factorizations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import TypeAlias

from math import sqrt
from operator import mul

from functools import reduce
from collections import defaultdict


Factorization : TypeAlias = dict[int, int] # dictionary of factors and their exponents

def is_prime(n : int) -> bool: # faster (but slightly less pretty) implementation
    '''Check if an integer is prime'''
    if n < 2:
        return False

    i = 2
    while (i * i) <= n:
        if (n % i) == 0:
            return False
        i += 1
    return True # must be prime if no smaller divisors are found

def is_prime_alt(n : int) -> bool: # more readable and Pythonic but slightly slower (same asymptotic complexity however)
    '''Check if an integer is prime'''
    if n < 2:
        return False

    for i in range(2, int(sqrt(n)) + 1):
        if (n % i) == 0:
            return False
    else:
        return True # must be prime if no smaller divisors are found

def prime_factorization(n : int) -> Factorization:
    '''Computes prime factorization of an integer n. Returns factorization as a dict,
    where the keys are the prime factors and the values are their respective exponents'''
    factors = defaultdict(int)
    
    i = 2
    while (i * i) <= n:
        while (n % i) == 0:
            n /= i
            factors[i] += 1
        i += 1

    if n > 2:
        factors[int(n)] = 1 # add any leftover large prime factors after division
    return factors

def num_from_factorization(fac : Factorization) -> int:
    '''Reconstruct a number as a composition of its factors'''
    return reduce(mul, (base**exp for base, exp in fac.items()))


'''Utilities and type-hinting for creating symbol inventories, which map symbols and bin labels to occurences
Useful concrete data structure for representing ordered sequences of symbol multisets for generalized ransom note enumeration'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Any, Generator, Generic, Iterable, Mapping, Optional, Sequence, TypeVar, Union

T = TypeVar('T')
L = TypeVar('L')

from importlib.util import find_spec
from functools import cached_property
from collections import Counter, defaultdict

try:
    import numpy as np
except ModuleNotFoundError:
    pass


def full_arr_builtin(*dims : Iterable[int], fill_value : Any=None) -> list[Any]:
    '''Similar to numpy.full (https://numpy.org/doc/stable/reference/generated/numpy.full.html) but with builtin list type'''
    arrlist = fill_value
    for dim_size in reversed(dims):
        if not isinstance(dim_size, int):
            raise TypeError(f'All dimensions must be of type int (not {dim_size.__class__.__name__})')
        arrlist = [arrlist for _ in range(dim_size)]
    return arrlist

class SymbolInventory(dict, Generic[T, L]):
    '''
    Representation of a map from a set of symbols of type T and a set of bin labels of type L to integer counts
    Implemented as a dict, keyed by symbol, whose values count the number of occurrences of that symbol in all bins that symbol occurs

    Data structure with specific methods which are useful in solving the generalized ransom note enumeration problem
    '''
    # INSTANTIATION
    def __init__(
        self, *args,
         _number_of_symbols : Optional[int]=None,
        _number_of_bins     : Optional[int]=None,
        _symbol_index_map   : Optional[dict[T, int]]=None,
        _bin_index_map      : Optional[dict[L, int]]=None,
        **kwargs,
    ) -> None:
        self._number_of_symbols = _number_of_symbols
        self._number_of_bins    = _number_of_bins

        self._symbol_index_map  = _symbol_index_map
        self._bin_index_map     = _bin_index_map

        super().__init__(*args, **kwargs)

    @classmethod
    def from_bin_sequence_mapping(cls,
        choice_bin_map : Mapping[L, Sequence[Iterable[T]]],
        _symbol_index_map : Optional[dict[T, int]]=None,
        _bin_index_map    : Optional[dict[L, int]]=None,
    ) -> 'SymbolInventory[T, L]':
        '''Initialize inventory from a mapping of labels to bins'''
        sym_inv_dict = defaultdict(Counter[L, int]) # keys are objects of type T ("symbols"), values give multiplicities of symbols keyed by bin position
        for label, choice_bin in choice_bin_map.items():
            for sym in choice_bin:
                sym_inv_dict[sym][label] += 1 # NOTE : implementation here requires that T be a hashable type
        
        return cls(
            sym_inv_dict,
            _number_of_symbols=len(sym_inv_dict), # pre-compute numbers of symbols and bins
            _number_of_bins =len(choice_bin_map), # pre-compute numbers of symbols and bins
            _symbol_index_map=_symbol_index_map, # enable iltering down of pre-computed maps from other initialization methods
            _bin_index_map=_bin_index_map,       # enable iltering down of pre-computed maps from other initialization methods
        )
    
    @classmethod
    def from_bin_sequence(cls, choice_bins : Sequence[Iterable[T]]) -> 'SymbolInventory[T, int]':
        '''Initialize inventory from an ordered sequence of bins
        Special case of mapping instantiation for sequences of bins; simply uses the index of a bin as its label'''

        choice_bin_map : dict[int, Sequence[Iterable[T]]] = {}
        _bin_index_map : dict[int, int] = {}
        for i, choice_bin in enumerate(choice_bins):
            choice_bin_map[i] = choice_bin # expand sequence into dict keyed by position in sequence 
            _bin_index_map[i] = i # ensures bin order does not get scrambled when initializing from this route (avoids confusion down the line)

        return cls.from_bin_sequence_mapping(
            choice_bin_map=choice_bin_map,
            _bin_index_map = _bin_index_map,
        )
    
    @classmethod
    def from_bins(cls, choice_bins : Union[Sequence[Iterable[T]], dict[L, Sequence[Iterable[T]]]]) -> 'SymbolInventory[T, L]':
        '''"Smart" initialization method which dispatches to from_bin_sequence() or from_bin_sequence_mapping(),
        depending on the nature of the "choice_bins" container provided'''
        if isinstance(choice_bins, Mapping):
            return cls.from_bin_sequence_mapping(choice_bin_map=choice_bins)
        elif isinstance(choice_bins, Sequence):
            return cls.from_bin_sequence(choice_bins=choice_bins)
        elif isinstance(choice_bins, Generator):
            return cls.from_bin_sequence(choice_bins=[i for i in choice_bins])
        else:
            raise TypeError(f'{cls.__name__} does not support initialization from non Sequence/Mapping container of type "{choice_bins.__class__.__name__}"')

    # INSTANCE METHODS
    def __repr__(self) -> str:
        '''Wrap dict repr to clarify that this is a subclass'''
        return f'{self.__class__.__name__}({super().__repr__()})'
    
    @property
    def number_of_symbols(self) -> int:
        '''Number of unique symbols present in the SymbolInventory'''
        if self._number_of_symbols is None:
            LOGGER.debug(f'Initializing field {"number_of_symbols"} of {self!r}')
            self._number_of_symbols = len(self.keys())
        return self._number_of_symbols
    num_symbols = num_sym = n_symbols = n_sym = number_of_symbols # aliases for convenience 

    @property
    def number_of_bins(self) -> int:
        '''Number of unique symbols present in the SymbolInventory'''
        if self._number_of_bins is None:
            LOGGER.debug(f'Initializing field {"number_of_bins"} of {self!r}')
            self._number_of_bins = self.involution.number_of_symbols
        return self._number_of_bins
    num_bins = num_bin = n_bins = n_bin = number_of_bins # aliases for convenience

    @property
    def symbol_index_map(self) -> dict[T, int]:
        '''Arbitrary one-to-one mapping between all symbols present in the inventory and integer indices'''
        if self._symbol_index_map is None:
            LOGGER.debug(f'Initializing field {"symbol_index_map"} of {self!r}')
            self._symbol_index_map = {
                symbol : i
                    for i, symbol in enumerate(self.keys())
            }
        return self._symbol_index_map

    @property
    def bin_index_map(self) -> dict[T, int]:
        '''Arbitrary one-to-one mapping between all bins present in the inventory and integer indices'''
        if self._bin_index_map is None:
            LOGGER.debug(f'Initializing field {"bin_index_map"} of {self!r}')
            self._bin_index_map = self.involution.symbol_index_map
        return self._bin_index_map
    
    def contains_word(self, sequence : Sequence[T], ignore_multiplicities : bool=False) -> bool:
        '''
        Check if a word (i.e. sequence of symbols of type T) could possibly be produced from a SymbolInventory

        Returns bool; True indicated the word can be made from the inventory
        False return indicates the sequence contains symbols not present in the inventory 
        OR contains more of one particular symbol than are present in the whole inventory
        '''
        for symbol, count in Counter(sequence).items():
            if symbol not in self:
                return False
            elif self[symbol].total() < (1 if ignore_multiplicities else count): # at least one occurence suffices if we don't care about exact counts
                return False # NOTE: could merge with above check (via OR operator short-circuit), but separated here for readability
        else:
            return True

    ## CONVERSION TO ALTERNATE FORMS
    def deepcopy(self) -> 'SymbolInventory[T, L]': # NOTE : deliberately named this "deepcopy" to reserve the name for a potetial cheaper shallow-copy "copy" method
        '''Create a deep copy of the current SymbolInventory'''
        return self.__class__(
            {symbol : bin_counts.copy() for symbol, bin_counts in self.items()}, # TOSELF : run some more tests to verify Counter.copy() does not point to original Counter
            _number_of_symbols=self._number_of_symbols,
            _number_of_bins=self._number_of_bins,
            _symbol_index_map=self._symbol_index_map,
            _bin_index_map=self._bin_index_map,
        )

    @cached_property
    def involution(self) -> 'SymbolInventory[L, T]':
        '''Returns a new SymbolInventory which switches the order of precedence of the symbols and bin labels
        So-called because self.involution.involution returns a SymbolInventory identical to self (by construction)'''
        invol = defaultdict(Counter[T, int])
        for symbol, bin_counts in self.items():
            for bin_label, count in bin_counts.items():
                invol[bin_label][symbol] += count
        return self.__class__(invol)
    
    @cached_property
    def occurence_matrix(self) -> Union[list[list[int]], np.ndarray[Any, int]]:
        '''
        Creates an NxM matrix (N is the number of symbols, and M is the number of bins) which represents a SymbolInventory
        Matrix element A_ij denotes the number of occurences of the i-th symbol (according to an arbitrary numbering) in the j-th bin

        Will attempt to return matrix as a 2-D numpy array, but will default to a doubly-nested list if numpy is not found to be installed
        '''
        shape : tuple[int, int] = (self.number_of_symbols, self.number_of_bins)
        if find_spec('numpy') is not None:
            occ_matr = np.zeros(shape, dtype=int)
        else: # default to list if numpy is not installed
            occ_matr = full_arr_builtin(*shape, fill_value=0)
        
        for symbol, bin_counts in self.items():
            i = self.symbol_index_map[symbol]
            for bin_label, count in bin_counts.items():
                j = self.bin_index_map[bin_label]
                occ_matr[i, j] = count
        
        return occ_matr


'''Unit-aware compendium of useful physical constants'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from scipy.constants import physical_constants
from .unitstr import unit_from_unit_str


PHYS_CONSTS_WITH_UNITS = {}
for key, (value, unit_str, uncert) in physical_constants.items():
    if (unit_val := unit_from_unit_str(unit_str)) is not None:
        PHYS_CONSTS_WITH_UNITS[key] = value * unit_val

# variable-names of physical constants from scipy.constants, which were in turn taken from NIST CODATA (https://physics.nist.gov/cuu/Constants/)
named_phys_consts = { 
    'c'         : ('speed of light in vacuum', 'speed_of_light'),
    'mu_0'      : ('vacuum mag. permeability', None),
    'epsilon_0' : ('vacuum electric permittivity', None),
    'h'         : ('Planck constant', 'Planck'),
    'hbar'      : ('Planck constant over 2 pi', None),
    'G'         : ('Newtonian constant of gravitation', 'gravitational_constant'),
    'g'         : ('standard acceleration of gravity', None),
    'e'         : ('elementary charge', 'elementary_charge'),
    'R'         : ('molar gas constant', 'gas_constant'),
    'alpha'     : ('fine-structure constant', 'fine_structure'),
    'N_A'       : ('Avogadro constant', 'Avogadro'),
    'k'         : ('Boltzmann constant', 'Boltzmann'),
    'sigma'     : ('Stefan-Boltzmann constant', 'Stefan_Boltzmann'),
    'Wien'      : ('Wien wavelength displacement law constant', None),
    'Rydberg'   : ('Rydberg constant', None),
    'm_e'       : ('electron mass', 'electron_mass'),
    'm_p'       : ('proton mass', 'proton_mass'),
    'm_n'       : ('neutron mass', 'neutron_mass'),
    'm_u'       : ('atomic mass constant', 'atomic_mass'),
}

for var_name, (CODATA_key, alias) in named_phys_consts.items():
    const_with_units = PHYS_CONSTS_WITH_UNITS[CODATA_key]
    globals()[var_name] = const_with_units # inject into global namespace
    if alias is not None:
        globals()[alias] = const_with_units


'''For refining a residue cover of an OpenFF Molecule into a partition, recording partition assignments to atom metadata.
 Must be called after successful call to openff.toolkit.topology.topology.from_pdb()
Code derived from original implementation and conceptualization by Connor Davel (https://github.com/openforcefield/polymer_examples/blob/main/monomer_generation/partition.py) '''

__author__ = 'Connor Davel, Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import TypeAlias

import json
from itertools import combinations
from collections import defaultdict

import networkx as nx
import numpy as np

from openff.toolkit.topology import Topology


# CUSTOM TYPEHINTS
BondInfoDict : TypeAlias = dict[tuple[int, int], int] # dictionary mapping bond start-and-end atom indices to the bond indices

# HELPER UTILITY FUNCTIONS
def _create_choice_graph(isomorphism_info : tuple[list[int], BondInfoDict]) -> nx.Graph:
    # TODO : add docstring description
    ''''''
    isomorphism_info = list(isomorphism_info.values()) # original ids are of no use
    all_values = set()
    for isomorphism in isomorphism_info:
        iso_ids, adjacencies = isomorphism
        for value in iso_ids:
            all_values.add(value)
    max_val = max(all_values)
    min_val = min(all_values)
    # create a matrix to find atom overlaps
    matrix = np.zeros((len(isomorphism_info), max_val + 1))
    adjacency_list = []
    for idx, isomorphism in enumerate(isomorphism_info):
        iso_ids, adjacencies = isomorphism
        cap_ids = []
        adjacencies_tupled = [] # adjacencies in tuple form for easier searching
        for bond, bond_type in adjacencies.items():
            edge_atom, cap_atom = bond
            cap_ids.append(cap_atom)
            adjacencies_tupled.append(tuple([edge_atom, cap_atom, bond_type]))
        adjacency_list.append(adjacencies_tupled)

        matrix[idx, list(set(iso_ids) - set(cap_ids))] = 1

    # matrix_data and adjacency_list are parallel to eachother
    # each item in adjacency_list corresponds to the isomorphism found in the 
    # corresponding row of matrix_data
    n_rows, n_cols = matrix.shape
    G = nx.Graph()
    all_bonds = []
    for iso_id, isomorphism in enumerate(adjacency_list):
        n_atoms = len(isomorphism_info[iso_id][0])
        if isomorphism == []:
            continue
        neighbors = []
        for bond in isomorphism:
            edge_a, cap_a, bond_type = bond
            all_bonds.append((edge_a, cap_a, bond_type, iso_id))

        row, = np.nonzero(np.squeeze(np.asarray(matrix[iso_id, :])))
        row = list(row)
        # find overlaps using matrix 
        overlaps, = np.nonzero(np.asarray(matrix[:, row].sum(axis=1)).reshape(-1))
        overlaps = list(overlaps)
        overlaps.remove(iso_id)
        G.add_node(
            iso_id,
            selected = False,
            overlaps = overlaps,
            bonds = isomorphism,
            n_atoms = n_atoms
        )
    for bond1, bond2 in combinations(all_bonds, 2):
        edge_a1, cap_a1, bond_type1, iso_id1 = bond1
        edge_a2, cap_a2, bond_type2, iso_id2 = bond2
        if (edge_a2, cap_a2, bond_type2) == (cap_a1, edge_a1, bond_type1):
            G.add_edge(
                iso_id1,
                iso_id2,
                order = bond_type1
            )

    return G

def _traverse(starting_queue : list[int], graph : nx.Graph) -> list[int]:
    # TODO : add docstring description
    found_nodes = []
    found_edges = []
    queue = starting_queue

    while len(queue) > 0:
        if queue == []: # TB : isn't this check redundant immediately after the while loop?
            return found_nodes
        
        v = queue.pop(0)
        found_nodes.append(v)
        for bond in graph.nodes[v]['bonds']:
            edge_a, cap_a, bond_type = bond
            if {edge_a, cap_a} in found_edges:
                continue

            selected_neighbor = -1
            for neighbor in graph.neighbors(v):
                # make sure this is the neighbor with the correct bond info
                neighbor_node = graph.nodes[neighbor]
                if (cap_a, edge_a, bond_type) not in neighbor_node['bonds']:
                    continue
                if neighbor_node['selected'] == True:
                    selected_neighbor = -1
                    break
                selected_neighbor = neighbor

            if selected_neighbor >= 0:
                queue.append(selected_neighbor)
                found_edges.append({edge_a, cap_a})

    return found_nodes

# MAIN PARTITIONING FUNCTION
def partition(offtop : Topology) -> bool:
    """
    args
    ----
    offtop: Openforcefield toolkit topology with complete metadata following successful call
        to openff.toolkit.topology.topology.from_pdb().
    returns 
    -------
    True if partition was successful
    False otherwise. This does not mean that the resulting assignment is invalid. It only means
        that clean partitioning cannot be achieved and some matches necessarily overlap. Try choosing
        other substructures if partitioning is necessary. 
    """
    for offmol in offtop.molecules:
        # first, populate isomorphism_info, which has the following format:
        #  [([target_ids], [{(map_start, map_end): bond_type}])]
        # isomorphism_info = dict()
        target_ids  = defaultdict(list)
        bond_info   = defaultdict(dict)
        match_names = defaultdict(str)

        for atom in offmol.atoms:
            iso_info = {
                int(k) : entry
                    for k, entry in json.loads(atom.metadata["match_info"]).items()
            }

            for query_num, query_name_id in iso_info.items():
                match_name = query_name_id[0] # id of the query atom
                query_id   = query_name_id[1] # total number of query 
                match_names[query_num] = match_name
                if atom.molecule_atom_index not in target_ids[query_num]:
                    target_ids[query_num].append(atom.molecule_atom_index)
                
                for b in atom.bonds:
                    # try to find inter-monomer bonds
                    if b.atom1_index == atom.molecule_atom_index:
                        begin_atom = b.atom1
                        end_atom = b.atom2
                    else:
                        begin_atom = b.atom2
                        end_atom = b.atom1
                        
                    if query_num in [int(k) for k in json.loads(end_atom.metadata["match_info"]).keys()]:
                        neighbor = False # TB : could make into explicit boolean?
                    else:
                        neighbor = True
        
                    if neighbor: # TB : could just perform above check
                        bond_entry = tuple([begin_atom.molecule_atom_index, end_atom.molecule_atom_index])
                        if bond_entry not in bond_info[query_num]:
                            bond_info[query_num][bond_entry] = b.bond_order
        
        if len(bond_info) == 0:
            continue 
        assert len(target_ids) == len(bond_info)

        isomorphism_info_dict = {idx: (target_ids[idx], bond_info[idx]) for idx in target_ids}
        idx_to_match_id = {idx: match_id for idx, match_id in enumerate(target_ids)}
        choice_G = _create_choice_graph(isomorphism_info_dict)
        connected_component_counts = []
        biggest_chain = None
        biggest_chain_length = 0
        for chain in nx.connected_components(choice_G):
            subgraph = choice_G.subgraph(chain)
            not_searched_nodes = list(subgraph.nodes)
            while len(not_searched_nodes) != 0:
                unique_group = _traverse([not_searched_nodes[0]], subgraph)
                new_tally = 0
                old_tally = 0
                overlapping_nodes = []
                for node in unique_group:
                    new_tally += subgraph.nodes[node]['n_atoms']
                    for overlapping_node in subgraph.nodes[node]['overlaps']:
                        if overlapping_node in subgraph.nodes and subgraph.nodes[overlapping_node]['selected'] == True:
                            overlapping_nodes.append(overlapping_node)
                            old_tally += subgraph.nodes[overlapping_node]['n_atoms']

                if new_tally > old_tally:
                    # exchange new for old choices 
                    for node in unique_group:
                        subgraph.nodes[node]['selected'] = True
                    for node in overlapping_nodes:
                        subgraph.nodes[node]['selected'] = False

                [not_searched_nodes.remove(i) for i in unique_group if i in not_searched_nodes] # TB : why is this a list comprehension?

            tally = 0
            for node in subgraph.nodes:
                if subgraph.nodes[node]['selected']:
                    tally += subgraph.nodes[node]['n_atoms']
            if tally > biggest_chain_length:
                biggest_chain = subgraph
                biggest_chain_length = tally
            # finally pick from unique_mapping_groups the largest list
        largest_mapping_group = []
        for node in biggest_chain.nodes:
            if biggest_chain.nodes[node]['selected']:
                largest_mapping_group.append(node)

        all_ids = []
        for iso_id in largest_mapping_group:
            all_ids += target_ids[idx_to_match_id[iso_id]]

        if set(all_ids) == set([a.molecule_atom_index for a in offmol.atoms]):
            # assign metadata
            for iso_id in largest_mapping_group:
                res_id = idx_to_match_id[iso_id]
                molecule_ids = target_ids[res_id]
                match_name = match_names[res_id]
                for m_id in molecule_ids:
                    atom = offmol.atom(m_id)
                    iso_info = {
                        int(k) : entry
                            for k,entry in json.loads(atom.metadata["match_info"]).items()
                    }
                    picked_match = iso_info[res_id]

                    # new metadata
                    atom.metadata["residue_name"] = picked_match[0]
                    atom.metadata["substructure_query_id"] = picked_match[1]
                    atom.metadata["residue_number"] = res_id
        else:
            return False # if one mol fails, return False
    return True # if all are successful, return True


'''Representation of vectors of coordinates and elementary distance geometry operations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generic, Optional, TypeVar, Union
from numbers import Number
from ...genutils.typetools.numpytypes import Shape, N, D

import numpy as np
from itertools import product as cartesian_product


# CLASSES
Num = TypeVar('Num', bound=Number) # geneic typehint for numeric coordinate values 
class Coordinates(Generic[Num]):
    '''Represention class for encapsulating N-vectors of D-dimensional coordinates, with basic coordinate operations'''
    def __init__(self, points : Optional[np.ndarray[Shape[N, D], Num]]=None) -> None:
        if points is None:
            points = np.array([[]])

        assert(points.ndim == 2) # only accepts vector of coordinate
        self.points = points
        self.n_dims = points.shape[-1] # the number of dimensions of a point in the vector

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.n_dims}-dimensional vector of {self.n_points} points)'

    # COORDINATE VECTOR DIMENSIONS AND EXTREMAL POINTS
    @property
    def n_points(self) -> int:
        '''The number of points currently stored in the coordinates vector'''
        return self.points.shape[0] # len(self.points)

    @property
    def dimensions(self) -> np.ndarray[Shape[D], Num]:
        '''The side lengths of the bounding box along each dimension'''
        return self.points.ptp(axis=0)
    dims = sidelens = sidelengths = dimensions

    @property
    def minimum(self) -> np.ndarray[Shape[D], Num]:
        '''The bounding box vertex with the smallest coordinates in each dimension'''
        return self.points.min(axis=0)
    min = lower = smallest = minimum

    @property
    def maximum(self) -> np.ndarray[Shape[D], Num]:
        '''The bounding box vertex with the largest coordinates in each dimension'''
        return self.points.max(axis=0)
    max = upper = largest = maximum

    @property
    def extrema(self) -> np.ndarray[Shape[2, D], Num]:
        '''A 2xN array of the minimal and maximal bounding box vertices'''
        return np.vstack([self.minimum, self.maximum])

    # POINT OPERATIONS
    def __call__(self, index : int) -> np.ndarray[Shape[D], Num]:
        '''Retrieve the point at the given index'''
        return self.points[index]

    def _point_is_compat(self, point : np.ndarray[Shape[D], Num]) -> bool:
        '''Internal method to check whether an incoming point is compatible with the dimensions of the current array'''
        return (
            isinstance(point, np.ndarray)
            and (point.ndim == 1)
            and (point.size == self.n_dims)
            # TODO: check compatible types? (may be too inflexible)
        )
    
    def validate_point(self, point : np.ndarray[Shape[D], Num]) -> None:
        '''Check if a point is compatible with the coordinates, or else raise Exception with reason why not'''
        if not self._point_is_compat(point):
            raise ValueError(f'Incompatible point {point}')

    def dists_to_point(self, point : np.ndarray[Shape[D], Num], norm_order : Optional[int]=None) -> np.ndarray[Shape[N], Num]:
        '''The distance between each point in a coordinate array and a single arbitrary point'''
        assert(self._point_is_compat(point))
        return np.linalg.norm(self.points - point, ord=norm_order, axis=1)

    # MEASURES OF CENTRALITY
    def weighted_centroid(self, weights : Optional[np.ndarray[Shape[N], Num]]=None) -> np.ndarray[Shape[D], Num]:
        '''The average (center-of-mass) coordinate of a vector of coordinates, with optional array of weights for each coordinates'''
        if weights is None:
            weights = np.ones(self.n_points, dtype=self.points.dtype)
        
        # prechecks to ensure the weights vector is of compatible size
        assert(weights.ndim == 1)
        assert(weights.size == self.points.shape[0])
        weights = weights.reshape((self.points.shape[0], 1))
        
        return (self.points * weights).mean(axis=0)

    @property
    def centroid(self) -> np.ndarray[Shape[D], Num]:
        '''The unweighted average coordinate of the vector'''
        return self.weighted_centroid() # weighed centroid with default unit weights
    center_of_mass = COM = centroid

    def dists_to_centroid(self, norm_order : Optional[int]=None, weights : Optional[np.ndarray[Shape[N], Num]]=None) -> np.ndarray[Shape[N], Num]:
        '''The distance of each coordinate in an array of coordinates to the coordinates' centroid'''
        return self.dists_to_point(point=self.weighted_centroid(weights=weights), norm_order=norm_order)
    radii = rad = dists_to_centroid

    # POINT ORDERINGS
    @property
    def lex_ordered_idxs(self) -> np.ndarray[Shape[N, D], int]: # TOSELF: "int" is the correct type annotation here (NOT Num)
        '''Returns a vector of the position that each point in self.points occupies when ordered lexicographically'''
        return np.lexsort(self.points.T)

    @property
    def lex_ordered_points(self) -> np.ndarray[Shape[N, D], Num]:
        '''Return copy of the points in the lattice in lexicographic order'''
        return self.points[self.lex_ordered_idxs]

    def lex_order_points(self) -> None:
        '''Sort points in the lattice in lexicographic order'''
        self.points = self.lex_ordered_points

    def randomize_points(self) -> None:
        '''Place the points in the lattice in a random order'''
        np.random.shuffle(self.points)

    # LATTICE TRANSFORMATIONS
    def translate(self, displacement : np.ndarray[Shape[D], Num]) -> None:
        '''Apply affine shift (translation only) to all points'''
        self.validate_point(displacement)
        self.points += displacement # TODO: use explicit broadcast here to reduce ambiguity?

    def linear_transformation(self, matrix : np.ndarray[Shape[D,D], float], as_coords : bool=False) -> Union[np.ndarray[Shape[N, D], float], 'Coordinates']:
        '''Accepts an NxN matrix (where D is the dimension of the lattice), returns a linearly-transformed copy of the coordinate points'''
        assert(matrix.shape == (self.n_dims, self.n_dims))
        transformed_points = self.points @ matrix.T # NOTE: need to right-multiply and transpose, since ROWS of self.points need to be tranformed 

        if as_coords:
            # return self.__class__(transformed_points) # TOSELF: this form is more general, but doesn't play nicely with inheritance 
            return Coordinates(transformed_points)
        return transformed_points

    def affine_transformation(self, matrix : np.ndarray[Shape[D,D], float], as_coords : bool=False) -> Union[np.ndarray[Shape[N, D], float], 'Coordinates']: # TOSELF: typehint on input matrix should be of shape D+1, D+1
        '''Accepts an NxN matrix (where D is the dimension of the lattice), returns an affinely-transformed copy of the coordinate points'''
        assert(matrix.shape == (self.n_dims + 1, self.n_dims + 1))
        aug_points = np.concatenate([self.points, np.ones((self.n_points, 1), dtype=int)], axis=1) # augment points vectors with extra columns of ones
        aug_transformed = aug_points @ matrix.T
        transformed_points = aug_transformed[: , :self.n_dims] / aug_transformed[:, self.n_dims, None] # downcast augmented transformed points from homogeneous coordinates, normalizing by projective part

        if as_coords:
            # return self.__class__(transformed_points) # TOSELF: this form is more general, but doesn't play nicely with inheritance 
            return Coordinates(transformed_points)
        return transformed_points


class BoundingBox(Coordinates):
    '''For representing a minimum bounding box around an N-coordinate vector of D-dimensional points'''
    def __init__(self, coords : Union[Coordinates, np.ndarray[Shape[N, D], Num]]) -> None:
        if isinstance(coords, np.ndarray):
            coords = Coordinates(coords) # allow passing of Coordinates-like classes
        points = np.array([vertex for vertex in cartesian_product(*coords.extrema.T)])
        
        super().__init__(points)

    @property
    def vertices(self) -> np.ndarray[Shape[N, D], Num]: # TOSELF : first axis of returned array is actually of size 2**D (haven't implemented typing yet)
        '''Alias of the "points" attribute to distinguish use via syntax'''
        return self.points
    
    @property
    def face_indices(self) -> np.ndarray[Shape[N, D], int]:
        '''Returns a [2**(n_dim-1) x 2*n_dim] whose points are the indices of (2*n_dims)-tuplets
        which form a "face" of the bounding box (i.e. shared values of exactly one coordinate)'''
        point_idxs = np.arange(self.n_points)
        step_mask = (self.points == self.minimum) # masking by max would work equally well (since all coordinates are shared...
        # step_mask = (bbox.points == bbox.maximum) #  ...with either the min or the max) and just yield a different order
        face_indices = []
        for i in range(self.n_dims):
            face_indices.append(point_idxs[ step_mask[:, i]])
            face_indices.append(point_idxs[~step_mask[:, i]])
        return np.array(face_indices) # TODO: account for correct order of rows and orientation of face within rows to give contiguous hypercube mesh
    
    @property
    def face_coords(self) -> np.ndarray[Shape[N, D, 3], int]:
        '''Return a [2**(n_dim-1) x 2*n_dim x 3] array in which each row is the set'''
        return self.points[self.face_indices]

    def surrounds(self, coords : Union[Coordinates, np.ndarray[Shape[N, D], Num]], strict : bool=False) -> np.ndarray[Shape[N, D], bool]:
        '''Boolean mask of whether the coordinates in a point vector lies within the bounding box'''
        if isinstance(coords, np.ndarray):
            coords = Coordinates(coords)
        assert(coords.n_dims == self.n_dims)
        less_funct = np.less if strict else np.less_equal # set "<" vs "<=" check by strict flag

        return less_funct(self.lower, coords.points) & less_funct(coords.points, self.upper)


'''Direct implementations of property calculations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import numpy as np
from ..maths.linearalg.decomposition import diagonalize


# RECOMMEND USING (SUCH AS MDTraj, MDAnalysis, ETC) FOR SPEED
def compute_gyration_tensor(coords : np.ndarray, use_eins=True) -> np.ndarray:
    '''Determines the gyration tensor of a set of atomic coordinates (and its diagonalization)
    Expects coordinates in an Nx3 array, with each column representing the x, y, and z coordinates, respectively'''
    N, ncols = arr_shape = coords.shape
    if ncols != 3:
        raise ValueError(f'Coordinate array must be of shape Nx3, not {arr_shape}')

    COM = coords.mean(axis=0) # centre-of-mass coord
    rs = coords - COM # place origin at center of mass

    if use_eins: # use Einstein summation for notational and memory compactness
        gyr_tens = np.einsum('ij, ik->jk', rs, rs) / N 
    else: # use direct displacement summation, less efficient but easier to understand
        gyr_tens = np.zeros((3, 3))
        for point in rs:
            S_i = np.outer(point, point)
            gyr_tens += S_i
        gyr_tens /= N

    return gyr_tens

def compute_Rg_and_K2(gyr_tens : np.ndarray, use_diag : bool=False) -> tuple[float, float]:
    '''Determine the radius of gyration (Rg) and radial shape anisotropy (K2) from a molecules gyration tensor'''
    if use_diag:
        P, D, P_inv = diagonalize(gyr_tens)
        I1 = np.trace(D)
        I2 = (I1**2 - np.trace(D**2)) / 2
    else:
        S, _eivecs = np.linalg.eig(gyr_tens)
        I1 = S.sum()
        I2 = (I1**2 - np.sum(S**2)) / 2

    Rg = np.sqrt(I1)
    K2 = 1 - 3*(I2 / I1**2)
    
    return Rg, K2


'''Utilities for handlling setup of Simulation Reporters'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Any, Callable, Optional, TypeAlias, Union
from dataclasses import dataclass, field
from pathlib import Path

from openmm.app import PDBReporter, PDBxReporter, DCDReporter
from openmm.app import StateDataReporter, CheckpointReporter

from ...genutils.typetools.parametric import Args, KWArgs
from ...genutils.fileutils.pathutils import assemble_path
from ...genutils.fileutils.jsonio.jsonify import make_jsonifiable
from ...genutils.fileutils.jsonio.serialize import PathSerializer


# REPORTER-SPECIFIC TYPES AND TYPEHINTS
class StateReporter(CheckpointReporter):
    '''CheckpointReporter which is forced to report States; greatly simplifies interface, documentation, and logging'''
    def __init__(self, file, reportInterval):
        super().__init__(file, reportInterval, writeState=True)

TrajectoryReporter  : TypeAlias = Union[PDBReporter, PDBxReporter, DCDReporter]
Reporter            : TypeAlias = Union[CheckpointReporter, StateReporter, StateDataReporter, TrajectoryReporter]
ReporterInitializer : TypeAlias = Callable[[Args, KWArgs], Reporter]


# REPORTER FILE EXTENSIONS AND INITIALIZERS
@dataclass(frozen=True)
class ReporterConfig:
    '''Interface for generating a Reporter and associated Path in a uniform and standardized way'''
    label     : str
    extension : str
    reporter_dict : dict[str, ReporterInitializer] # dictionary of intializers keyed by extension
    extra_kwargs  : dict[str, Any] = field(default_factory=dict)

    @property
    def initializer(self) -> ReporterInitializer:
        '''Returns the particular class or initializer'''
        return self.reporter_dict[self.extension] # explicitly called with getitem to raise KeyError if extension is invalid

TRAJ_REPORTERS = { # index output formats of reporters by file extension
    'pdb'  : PDBReporter,
    'pdbx' : PDBxReporter,
    'dcd'  : DCDReporter,
}

CHK_REPORTERS = {
    'chk' : CheckpointReporter, # TODO : consider adding checks to guarantee that this can't ever be called with writeState=True
    'xml' : StateReporter
}

STATE_DAT_REPORTERS = {
    'csv' : StateDataReporter
}


# SERIALIZABLE REPORTER PARAMETER STORAGE AND GENERATION
DEFAULT_STATE_DATA_PROPS = {
    'step'            : True,
    'time'            : True,
    'potentialEnergy' : True,
    'kineticEnergy'   : True,
    'totalEnergy'     : True,
    'temperature'     : True,
    'volume'          : True,
    'density'         : True,
    'speed'           : True,
    'progress'        : False,
    'remainingTime'   : False,
    'elapsedTime'     : False
}

@make_jsonifiable(type_serializer=PathSerializer)
@dataclass
class ReporterParameters:
    '''Parameters for specifying Simulation reporters'''
    report_checkpoint : bool = True
    report_state      : bool = True
    report_trajectory : bool = True
    report_state_data : bool = True

    traj_ext   : Optional[str] = 'dcd'
    num_steps  : Optional[int] = None 
    state_data     : Optional[dict[str, bool]] = field(default_factory=lambda : DEFAULT_STATE_DATA_PROPS)
    reporter_paths : Optional[dict[str, Path]] = field(default=None)

    @property
    def rep_configs(self) -> list[ReporterConfig]:
        '''Generate Reporter configurations for currently specified reporters'''
        rep_configs = []
        
        if self.report_trajectory:
            rep_configs.append(ReporterConfig(label='trajectory', extension=self.traj_ext, reporter_dict=TRAJ_REPORTERS))

        if self.report_checkpoint:
            rep_configs.append(ReporterConfig(label='checkpoint', extension='chk', reporter_dict=CHK_REPORTERS))

        if self.report_state:
            rep_configs.append(ReporterConfig(label='state', extension='xml', reporter_dict=CHK_REPORTERS))

        if self.report_state_data:
            rep_configs.append(ReporterConfig(label='state_data', extension='csv', reporter_dict=STATE_DAT_REPORTERS, extra_kwargs={**self.state_data, 'totalSteps' : self.num_steps}))
                    
        return rep_configs

    def init_reporter_paths(self, out_dir : Path, prefix : str) -> None: # TODO : add mechanism to automatically re-initialize when bool params change (might onvolve caching dir info)
        '''Generate paths for all configured and enabled reporters'''
        self.reporter_paths = {
            rep_config.label : assemble_path(out_dir, prefix, extension=rep_config.extension, postfix=rep_config.label)
                for rep_config in self.rep_configs
        }

    def prepare_reporters(self, report_interval : int) -> list[Reporter]:
        '''Prepare all reporters specified by internal parameters and return Paths linked to said reporters'''
        if self.reporter_paths is None:
            raise ValueError('Cannot create Simulation reporters without initializing reporter output paths')
        
        reporters = []
        for rep_config in self.rep_configs:
            rep_path = self.reporter_paths[rep_config.label] # will raise KeyErro if config changes between Path init
            rep_path.touch() # ensure the reporter output file actually exists

            rep = rep_config.initializer(str(rep_path), reportInterval=report_interval, **rep_config.extra_kwargs)
            LOGGER.info(f'Prepared {rep.__class__.__name__} which reports to {rep_path}')
            reporters.append(rep)

        return reporters


'''Tools for replacing Ports and functional groups'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Optional

from rdkit import Chem
from rdkit.Chem import Mol, RWMol

from ._bonding import combined_rdmol
from .formation import increase_bond_order
from .portlib import get_linker_ids, get_single_port
from .identification import get_num_bondable_port_pairs, get_first_bondable_port_pair

from ..labeling import molwise
from ...genutils.decorators.functional import optional_in_place


@optional_in_place
def splice_atoms(rwmol : RWMol, atom_id_1 : Optional[int]=None, atom_id_2 : Optional[int]=None, flavor_pair : tuple[Optional[int], Optional[int]]=(None, None)) -> None:
    '''Completely combine a single pair of ports on two target atoms to create a bond of desired order'''
    port_1, port_2 = get_first_bondable_port_pair(rwmol, atom_id_1=atom_id_1, atom_id_2=atom_id_2, flavor_pair=flavor_pair) # TODO : reimplement this to be slightly less redundant (maybe add option to specify bond order ahead of time in increase_bond_order()?)
    if atom_id_1 is None: # fill in atom indices to give greater specificity for single bond order upconversion
        atom_id_1 = port_1.bridgehead.GetIdx()
    if atom_id_2 is None: # fill in atom indices to give greater specificity for single bond order upconversion
        atom_id_2 = port_2.bridgehead.GetIdx()

    assert(port_1.bond.GetBondType() == port_2.bond.GetBondType()) # slightly redundant given pre-bonding checks, but is a helpful failsafe
    for i in range(int(port_1.bond.GetBondTypeAsDouble())): # bond the target atoms up to the degree of the desired port pair; types
        if i == 0: # record map numbers, as these are invariant between bonding events (unlike atom IDs)
            atom_map_nums = [j for j in molwise.map_nums_by_atom_ids(rwmol, atom_id_1, atom_id_2)] # unpack as list to avoid values being "used up" upon iteration
        else:
            atom_id_1, atom_id_2 = molwise.atom_ids_by_map_nums(rwmol, *atom_map_nums) # reassign bonded atom IDs from invariant map nums cached on first bond formation

        increase_bond_order(rwmol, atom_id_1, atom_id_2, flavor_pair=flavor_pair, in_place=True) 

def saturate_ports(rdmol : Mol, cap : Mol=Chem.MolFromSmiles('[*]-[H]'), flavor_to_saturate : int=0) -> None:
    '''Takes an RDKit Mol and another "cap" molecule (by default just hydrogen) and caps all valid ports (with the specified flavor) on the target Mol with the cap group'''
    flavor_pair : tuple[int, int] = (flavor_to_saturate, get_single_port(cap).flavor) # will raise exception if cap has anything other than 1 port
    
    rwmol = combined_rdmol(rdmol, cap, editable=True) # create initial combination to test if bonding is possible
    num_bonds_formable = get_num_bondable_port_pairs(rwmol, flavor_pair=flavor_pair)
    if num_bonds_formable == 0:
        return rdmol # special case, if no bondable ports exist to begin with, return the original molecule
    
    # implicit else if initial return in null case is not hit
    while num_bonds_formable > 0: # not greater than 1, since initial combination is already done outside loop
        splice_atoms(rwmol, flavor_pair=flavor_pair, in_place=True)
        if num_bonds_formable > 1: # need -1 to exclude final addition
            rwmol = combined_rdmol(rwmol, cap, editable=True) # add another cap group on all but the final iteration. NOTE : must be added one-at-a-time to preclude caps from bonding to each other
        num_bonds_formable = get_num_bondable_port_pairs(rwmol, flavor_pair=flavor_pair) # update count of available bonds (this may change as new bonds are added)
    molwise.assign_ordered_atom_map_nums(rwmol, in_place=True) # ensure map numbers are ordered and minial
    
    return Chem.Mol(rwmol) # revert to "regular" Mol from RWMol

@optional_in_place # temporarily placed here for backwards-compatibility reasons
def hydrogenate_rdmol_ports(rdmol : Mol) -> None:
    '''Replace all port atoms with hydrogens'''
    # return Chem.ReplaceSubstructs(rdmol, Chem.MolFromSmarts('[#0]'), Chem.MolFromSmarts('[#1]'), replaceAll=True)[0] # changes order of atom ids, undesiredly
    for port_id in get_linker_ids(rdmol):
        rdmol.GetAtomWithIdx(port_id).SetAtomicNum(1)


'''Type-hinting for Union-bound categories of types'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Callable, Container, Iterable, Sequence, Type, TypeAlias, Union
from numpy import (
    ndarray,
    number as np_number
)
from numpy.typing import ArrayLike

import builtins
from inspect import isclass
from .parametric import U


# TYPECHECKING FOR CUSTOM UNION TYPE ALIASES
def _union_member_factory(union : U, regname : str='Union') -> Callable[[Any], bool]:
    '''Factory for making Union-membership-checking functions'''
    def isinunion(var : Any) -> bool:
        return isinstance(var, union.__args__)
    
    isunion_function_name = f'is{regname.lower()}'
    isinunion.__name__ = isunion_function_name
    isinunion.__qualname__ = isunion_function_name
    isinunion.__doc__ = f'''Check if an object is an instance of the {regname} class'''

    return isinunion

_UNION_TYPES : dict[str, U] = { # registry of aliases type Unions (will be dynamically registered with isinstance-like checkers at module level)
    'StringLike'        : Union[str, bytes, bytearray],                         # classes which 
    'ListLike'          : Union[list, tuple, set, frozenset, bytes, bytearray], # classes which are Container-like and can be initialized from a list of values
    'JSONSerializable'  : Union[str, bool, int, float, tuple, list, dict],      # classes which can be written to a JSON file by the default JSONEncoder/Decoder
}
for union_name, union_type in _UNION_TYPES.items():
    globals()[union_name] = union_type
    _union_checker_func = _union_member_factory(union_type, union_name) # register to module-level scope
    globals()[_union_checker_func.__name__] = _union_checker_func # register to module-level scope


# REGISTRIES OF ALL BUILTIN TYPES WITH GIVEN BASE CLASS BEHAVIOR
BUILTIN_TYPES : dict[str, Type] = {}
_BUILTIN_BASES_TO_CHECK : tuple[Type] = (Exception, Callable, Container, Iterable, Sequence)
for _klass in _BUILTIN_BASES_TO_CHECK:
    class_reg : dict[str, Type] = {}
    for builtin_name in dir(builtins):
        builtin_obj = getattr(builtins, builtin_name)
        if isclass(builtin_obj):
            BUILTIN_TYPES[builtin_obj.__name__] = builtin_obj
            if issubclass(builtin_obj, _klass):
                class_reg[builtin_obj.__name__] = builtin_obj
    globals()[f'BUILTIN_{_klass.__name__.upper()}S']  = class_reg # register to module-level scope


'''For obtaining info from and for labelling individual RDKit Bonds'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Callable, Iterable, Optional
from itertools import combinations

from rdkit import Chem
from rdkit.Chem.rdchem import Bond, BondType, Mol

from .molwise import atom_ids_by_map_nums
from ...genutils.typetools.parametric import Args, KWArgs
from ...genutils.iteration import sliding_window
from ...genutils.decorators.functional import optional_in_place


# BOND ID QUERYING
def are_bonded_atoms(rdmol : Mol, atom_id_1 : int, atom_id_2 : int) -> bool:
    '''Check if pair of atoms in an RDKit Mol have a bond between them'''
    return (rdmol.GetBondBetweenAtoms(atom_id_1, atom_id_2) is not None)

def get_bonded_pairs(rdmol : Mol, *atom_ids : Iterable[int]) -> dict[int, tuple[int, int]]:
    '''Get bond and terminal atom indices of all bonds which exist between any pair of atoms in an indexed list'''
    res = {}
    
    atom_id_pairs = combinations(atom_ids, 2)
    for atom_id_pair in atom_id_pairs:
        bond = rdmol.GetBondBetweenAtoms(*atom_id_pair)
        if bond is not None:
            res[bond.GetIdx()] = atom_id_pair
    return res

def bond_ids_by_cond(rdmol : Mol, bond_cond : Callable[[Bond, Args, KWArgs], bool]) -> tuple[int]:
    '''Return IDs of all bonds which satisfy some binary condition'''
    return tuple(
        bond.GetIdx()
            for bond in rdmol.GetBonds()
                 if bond_cond(bond)
    )

def get_bonded_pairs_by_map_nums(rdmol : Mol, *atom_map_nums : Iterable[int]) -> dict[int, tuple[int, int]]:
    '''Obtain bonded pair dict by atom map numbers instead of IDs'''
    return get_bonded_pairs(rdmol, *atom_ids_by_map_nums(rdmol, *atom_map_nums))

def get_bond_by_map_num_pair(rdmol : Mol, map_num_pair : tuple[int, int]) -> Bond:
    '''Get the index of a bond spanning a pair of atoms with given pair of atom map numbers'''
    return rdmol.GetBondBetweenAtoms(*atom_ids_by_map_nums(rdmol, *map_num_pair))

def get_bond_id_by_map_num_pair(rdmol : Mol, map_num_pair : tuple[int, int]) -> Bond:
    '''Get the index of a bond spanning a pair of atoms with given pair of atom map numbers'''
    return get_bond_by_map_num_pair(rdmol, map_num_pair).GetIdx()

def get_shortest_path_bonds(rdmol : Mol, start_idx : int, end_idx : int) -> list[int]:
    '''Returns bond indices along shortest path between two atoms in a Mol'''
    return [
        rdmol.GetBondBetweenAtoms(*atom_id_pair).GetIdx()
            for atom_id_pair in sliding_window(Chem.GetShortestPath(rdmol, start_idx, end_idx), n=2)
    ]


# BOND ID LABELING
@optional_in_place
def assign_bond_id_labels(rdmol : Mol, bond_id_remap : Optional[dict[int, int]]=None) -> None:
    '''Draws bond indices over their positions when displaying a Mol. 
    Can optionally provide a dict mapping bond indices to some other integers'''
    if bond_id_remap is None:
        bond_id_remap = {} # avoid mutable default

    for bond in rdmol.GetBonds():
        bond.SetIntProp('bondNote', bond_id_remap.get(bond.GetIdx(), bond.GetIdx())) # check if map value exists, if not default to index

@optional_in_place
def clear_bond_id_labels(rdmol : Mol) -> None:
    '''Removes bond indices over their positions when displaying a Mol'''
    for bond in rdmol.GetBonds():
        bond.ClearProp('bondNote')


'''Utilities for serializing, converting, and extracting info from OpenFF topologies'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from pathlib import Path
from ast import literal_eval
import numpy as np

from rdkit import Chem
from rdkit.Chem.rdchem import (
    Atom as RDAtom,
    Mol,
)

from openff.toolkit import ToolkitRegistry
from openff.toolkit.topology import (
    Atom as OFFAtom,
    Molecule,
    Topology,
)

from . import GTR
from ...genutils.iteration import asiterable
from ...genutils.fileutils.pathutils import dotless
from ...genutils.decorators.functional import allow_string_paths, optional_in_place

from ...rdutils.rdprops import RDPROP_GETTERS, RDPROP_SETTERS, copy_rd_props
from ...rdutils.rdcoords.tiling import tile_lattice_with_rdmol


# TOPOLOGY INFO
def get_largest_offmol(offtop : Topology) -> Molecule:
    '''Return the largest molecule in a Topology'''
    return max(offtop.molecules, key=lambda offmol : offmol.n_atoms)


# RDKIT METADATA-PRESERVING INTERCONVERSION
def copy_atom_metadata(offatom : OFFAtom, rdatom : RDAtom, preserve_type : bool=True) -> None:
    '''Copies all attributes from the metadata dict of an OpenFF-type Atom as Props of an RDKit-type atom'''

    for key, value in offatom.metadata.items():
        if (type(value) not in RDPROP_SETTERS) or (not preserve_type): # set as string if type is unspecified or if explicitly requested to
            rdatom.SetProp(key, str(value))
        else:
            setter = getattr(rdatom, RDPROP_SETTERS[type(value)]) # use the atom's setter for the appropriate type
            setter(key, value)

def to_rdkit_with_metadata(offmol : Molecule, preserve_type : bool=True) -> Mol:
    '''Converts an OpenFF molecule to an RDKit molecule, preserving atomic metadata'''
    rdmol = offmol.to_rdkit()
    for i, offatom in enumerate(offmol.atoms):
        copy_atom_metadata(offatom, rdmol.GetAtomWithIdx(i), preserve_type=preserve_type)

    return rdmol


# MOLECULE METADATA SERIALIZATION / DESERIALIZATION
@optional_in_place
def package_atom_metadata(offmol : Molecule) -> None:
    '''Collect atom metadata into a serializable property of the parent Molecule'''
    packaged_mdat = {
        atom.molecule_atom_index : atom.metadata
            for atom in offmol.atoms
                if atom.metadata
    }

    if packaged_mdat: # only assign if metadata is present (cleaner than assigning empty dict)
        offmol.properties['metadata'] = packaged_mdat

@optional_in_place
def unpackage_atom_metadata(offmol : Molecule) -> None:
    '''Reassign atom metadata from a "packaged" metadata property dict belonging to the parent Molecule'''
    packaged_mdat = offmol.properties.pop('metadata', None) # lookup and remove serialized data, or None if not present
    if not packaged_mdat: # handles both the null (i.e. NoneType) case and the empty dict case
        return
    
    if isinstance(packaged_mdat, str):
        packaged_mdat = literal_eval(packaged_mdat) # de-stringify if necessary

    for atom_id, metadata in packaged_mdat.items():
        offmol.atoms[atom_id].metadata.update(metadata)


# FILE I/O FUNCTIONS
@allow_string_paths
def save_molecule(path : Path, offmol : Molecule, toolkit_registry : ToolkitRegistry=GTR) -> None: # TODO : generalize to work for bytes w/ opened file
    '''Syntactic sugar for annoying suffix re-specification when saving OpenFF Molecules'''
    offmol.to_file(str(path), file_format=dotless(path), toolkit_registry=toolkit_registry)

@allow_string_paths
def topology_to_sdf(path : Path, offtop : Topology, toolkit_registry : ToolkitRegistry=GTR) -> None:
    '''Save an OpenFF Topology to an SDF file file atom metadata preserved'''
    assert(path.suffix == '.sdf')

    with path.open('w') as sdf_file:# TODO : maybe change to append mode instead?
        for mol in offtop.molecules:
            serial_mol = package_atom_metadata(mol, in_place=False)  # packageage metadata for serialization WITHOUT disturbing the original molecule
            serial_mol.to_file(sdf_file, file_format='SDF', toolkit_registry=toolkit_registry) # NOTE : not using save_molecule() here, as that function does not currently support bytes-like input (needed for multiple mols in same file)
    LOGGER.debug(f'Successfully serialized SDF Topology to {path}')

@allow_string_paths
def topology_from_sdf(path : Path, *args, **kwargs) -> Topology:
    '''Load an OpenFF Topology from an SDF file, assigning metadata and partial charges if stored'''
    assert(path.suffix == '.sdf')

    LOGGER.debug(f'Loading serialized SDF Topology from {path}')
    return Topology.from_molecules(
        unpackage_atom_metadata(mol, in_place=False) # reassign metadata if serialized
            for mol in asiterable(Molecule.from_file(path, *args, **kwargs)) # asiterable() gracefully handles the singleton case
    )


# TOPOLOGY BUILD FUNCTIONS
def topology_from_molecule_onto_lattice(offmol : Molecule, lattice_points : np.ndarray, rotate_randomly : bool=True, unique_mol_ids : bool=True):
    '''Convert a charged OpenFF Molecule into a Topology made up of copies of that Molecule tiled according to a lattice'''
    tiled_rdmol = tile_lattice_with_rdmol(offmol.to_rdkit(), lattice_points, rotate_randomly=rotate_randomly)
    tiled_offmols = [] 
    for mol_id, tiled_mol_copy in enumerate(Chem.GetMolFrags(tiled_rdmol, asMols=True, sanitizeFrags=False)):
        copy_rd_props(tiled_rdmol, tiled_mol_copy) # ensure each individual fragment preserves the information of the parent molecule
        tiled_offmol = Molecule.from_rdkit(
            rdmol=tiled_mol_copy,
            allow_undefined_stereo=True,
            hydrogens_are_explicit=True
        )
        if unique_mol_ids:
            for atom in tiled_offmol.atoms:
                atom.metadata['residue_number'] = mol_id # for now, this appears to be the mechanism by which Interchange assigns labels to unique molecules
        tiled_offmols.append(tiled_offmol)

    return Topology.from_molecules(tiled_offmols)


'''For bitwise operations and conversions to/from bitstrings'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union 


def int_to_bits(n : int, num_bits : int=8, clamp : bool=False, as_list : bool=False) -> Union[str, list[int]]:
    '''
    Convert an integer into a string of its bits, padded out to <num_bits> bits
    If clamp=True and the binary representation of <n> has more than <num_bits> bits, then leading bits will be discarded
    If as_list=True, will return as list of 0/1 ints; otherwise, will return as string
    '''
    bitstring = f'{n:0{num_bits}b}' # format into given number of bits with leading zeros - way easier than directly implementing bitshift-based operations
    if clamp:
        bitstring = bitstring[-num_bits:]

    if as_list:
        return [int(bit) for bit in bitstring] 
    return bitstring


'''Representation classes for monomer graphs bonds which encode intermonomer Port information'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import re
from rdkit import Chem

from typing import ClassVar, Union, Optional
from dataclasses import dataclass, field

from ...smileslib.primitives import BONDTYPE_BY_BOND_SMARTS, BOND_SMARTS_BY_BONDTYPE, BOND_PRIMITIVES_FOR_REGEX


# PROCESSING BOND TOKENS IN SMIDGE STRINGS
BOND_TOKEN_RE = re.compile(
    r'(?P<incoming_flavor>\d*)' \
    f'(?P<bondtype>{BOND_PRIMITIVES_FOR_REGEX})' \
    r'(?P<outgoing_flavor>\d*)'
)

@dataclass
class MonomerGraphBondInfo:
    '''Encapsulated information about an intermonomer bond in a monomer graph'''
    DEFAULT_BONDTYPE : ClassVar[Chem.BondType] = Chem.BondType.UNSPECIFIED
    DEFAULT_FLAVOR   : ClassVar[int] = 0

    incoming_flavor : int    = field(default=DEFAULT_FLAVOR)
    bondtype : Chem.BondType = field(default=DEFAULT_BONDTYPE)
    outgoing_flavor : int    = field(default=DEFAULT_FLAVOR)

    def __post_init__(self) -> None:
        '''Perform appropriate type conversions and apply defaults'''
        if self.incoming_flavor is None:
            self.incoming_flavor = self.DEFAULT_FLAVOR

        if self.bondtype is None:
            self.bondtype = self.DEFAULT_BONDTYPE

        if self.outgoing_flavor is None:
            self.outgoing_flavor = self.DEFAULT_FLAVOR

    @property
    def bond_str(self) -> str:
        '''SMARTS representation of the current bondtype (defaults to the default symbol if None or invalid bondtype is provided)'''
        return BOND_SMARTS_BY_BONDTYPE.get(self.bondtype, BOND_SMARTS_BY_BONDTYPE.get(self.DEFAULT_BONDTYPE))

    def __str__(self) -> str:
        return f'{self.incoming_flavor or ""}{self.bond_str}{self.outgoing_flavor or ""}'

    
    @staticmethod
    def parse_str_dict(str_dict : dict[str, str]) -> dict[str, Optional[Union[str, Chem.BondType]]]:
        '''Parse string-valued dict into dict of correct types and NoneType defaults for empty keys'''
        # 0) create copy of dict to avoid any in-place modification
        imb_info = {k : v for k, v in str_dict.items()}
        
        # 1) process bond type
        imb_info['bondtype'] = BONDTYPE_BY_BOND_SMARTS.get(imb_info.get('bondtype')) # set to None either if no bondtype is provided OR the provided type is not a registered primitive

        # 2) process port flavors
        for flavor_attr in ('incoming_flavor', 'outgoing_flavor'):
            if not (flavor_str := imb_info.get(flavor_attr)): # raised when the flavor_str is either not present or empty
                imb_info[flavor_attr] = None
            elif isinstance(flavor_str, str) and flavor_str.isdigit():
                imb_info[flavor_attr] = int(flavor_str) # convert parsed strings to ints where possible

        return imb_info

    @classmethod
    def from_dict(cls, str_dict : dict[str, str]) -> 'MonomerGraphBondInfo':
        '''Initialize from dictionary of values, after sanitizing'''
        return cls(**cls.parse_str_dict(str_dict))

    @classmethod
    def from_match(cls, match : Optional[re.Match]) -> 'MonomerGraphBondInfo':
        '''Initialize from groupdict of regex Match'''
        if match is None:
            # TODO: add logged warning
            return cls()
        return cls.from_dict(match.groupdict())


"""
Unit and regression test for the polymerist package.
"""

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

# Import package, test suite, and other packages as needed
import sys

import pytest

import polymerist


def test_polymerist_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "polymerist" in sys.modules



'''For packing solvents into Topology boxes using packmol'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging

LOGGER = logging.getLogger(__name__)

from typing import Optional, Union

import numpy as np
from openmm.unit import Quantity

from openff.toolkit import Topology, Molecule, ForceField
from openff.interchange.components import _packmol as packmol

from . import physprops
from .. import boxvectors
from ..topology import get_largest_offmol


def pack_topology_with_solvent(offtop : Topology, solvent : Molecule, box_vecs : Union[boxvectors.VectorQuantity, boxvectors.BoxVectorsQuantity], density : Quantity, exclusion : Optional[Quantity]=None) -> Topology:
    '''Pack a Topology with a given solvent molecule to a given box size with desired density
    The dimensions of the desired box_dimensions '''
    # obtain bounding box (plus exclusion) for topology
    box_vecs = boxvectors.box_vectors_flexible(box_vecs) # up-convert to full box vectors if only dimensions are given
    min_box_vecs = boxvectors.get_topology_bbox(offtop)
    if exclusion is not None:
        min_box_vecs = boxvectors.pad_box_vectors_uniform(min_box_vecs, exclusion)

    if not np.all(box_vecs >= min_box_vecs): # TODO : change this to only check XYZ dimensions
        raise boxvectors.BoxVectorError(f'Desired box dimensions ({box_vecs}) are smaller than minimum excluded Topology dimensions ({min_box_vecs})')

    # assign properties from desired box vectors if size checks pass
    box_vol = boxvectors.get_box_volume(box_vecs, units_as_openm=True)
    N = physprops.num_mols_in_box(solvent, box_vol=box_vol, density=density)

    # packing waters into topology and saving
    LOGGER.info(f'Solvating {box_vol} Topology with {N} {solvent.name} molecules to density of {density}')
    packed_top = packmol.pack_box(
        [solvent],
        [N],
        offtop,
        box_vectors=box_vecs, 
        box_shape=packmol.UNIT_CUBE,
        center_solute='BRICK'
    )
    LOGGER.info('Packmol packing converged')

    packed_top.box_vectors = box_vecs
    LOGGER.info(f'Set solvated Topology box vectors to {box_vecs}')
    offmol = get_largest_offmol(packed_top)
    offmol.properties['solvent'] = solvent.name

    return packed_top


'''Interfaces between residue-charge calculation methods and OpenFF (or other external) tools'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass

from openff.toolkit import Molecule

from .rctypes import ChargesByResidue
from .calculation import apply_residue_charges
from ..molchargers import MolCharger
from .....genutils.decorators.functional import optional_in_place


@dataclass
class LibraryCharger(MolCharger, CHARGING_METHOD='RCT'):
    '''Charger class for applying library charges onto residue-mapped Molecules'''
    charges_by_residue : ChargesByResidue

    @optional_in_place
    def _charge_molecule(self, uncharged_mol : Molecule) -> None:
        apply_residue_charges(uncharged_mol, self.charges_by_residue, in_place=True) # TOSELF: class is defined at runtime, so the fact that this functions is defined later in the module doesn't really matter
        


'''Unit tests for PDB file I/O utils'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import pytest
from polymerist.molfiles.pdb import SerialAtomLabeller


ELEMS : tuple[str] = ('C', 'H', 'H', 'H', 'N', 'H', 'C', 'O', 'Cl') # atoms for methylcarbamoyl chloride (MCC)

@pytest.mark.parametrize(
    'mol_atom_elems, atom_label_width, include_elem_idx, default_elem_idx, expected_labels',
    [
        (ELEMS, 4, True, 0, ['C000', 'H000', 'H001', 'H002', 'N000', 'H003', 'C001', 'O000', 'Cl00']), # test with default PDD-compatible settings
        (ELEMS, 4, True, 2, ['C002', 'H002', 'H003', 'H004', 'N002', 'H005', 'C003', 'O002', 'Cl02']), # test element index offset
        (ELEMS, 3, True, 2, ['C02', 'H02', 'H03', 'H04', 'N02', 'H05', 'C03', 'O02', 'Cl2']), # test shorter atom label width
        (ELEMS, 1, True, 0, ['C', 'H', 'H', 'H', 'N', 'H', 'C', 'O', 'C']), # test truncation works below threshold where indices can be written
        (ELEMS, 4, False, 0, ['C   ', 'H   ', 'H   ', 'H   ', 'N   ', 'H   ', 'C   ', 'O   ', 'Cl  ']), # test without element indices
        (ELEMS, 4, False, 7, ['C   ', 'H   ', 'H   ', 'H   ', 'N   ', 'H   ', 'C   ', 'O   ', 'Cl  ']), # test that default indices has no impact when indices aren't present
        (ELEMS, 0, False, 0, ['', '', '', '', '', '', '', '', '']), # test null-width labels
        # Invalid input handling checks
        pytest.param(
            ELEMS, -1, True, 0, [], # test that negative label width is rejected as intended
            marks=pytest.mark.xfail(
                raises=ValueError,
                reason='Negative atom label widths not allowed',
                strict=True,
            )
        ),
        pytest.param(
            ELEMS, 4, True, -5, [], # test that negative default indices are rejected as intended
            marks=pytest.mark.xfail(
                raises=ValueError,
                reason='Negative element indices not allowed',
                strict=True,
            )
        ),
        pytest.param(
            tuple(len(elem) for elem in ELEMS), 4, True, 0, [], # test that negative default indices are rejected as intended
            marks=pytest.mark.xfail(
                raises=TypeError,
                reason='Must pass atom elements as strings',
                strict=True,
            )
        ),
    ]
)
def test_atom_labeller(
        mol_atom_elems : tuple[str],
        atom_label_width : int,
        include_elem_idx : bool,
        default_elem_idx : int,
        expected_labels : list[str],
    ) -> None:
    '''Test that atom labelling hebaves as expected with various label formatting configurations'''
    labeller = SerialAtomLabeller(
        atom_label_width=atom_label_width,
        include_elem_idx=include_elem_idx,
        default_elem_idx=default_elem_idx,
    )
    assert [labeller.get_atom_label(elem) for elem in mol_atom_elems] == expected_labels


'''For curating pre-defined solvent molecules'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from pathlib import Path # TODO : reimplement "properly" "using importlib_resources (more complicated than it's worth for now)
_MODULE_PATH = Path(__path__[0])

from openff.toolkit import Molecule
from openff.units import unit as offunit

from ... import topology


def generate_water_TIP3P() -> Molecule:
    '''Helper method for creating a new TIP3P water representation from scratch'''
    TIP3P_ATOM_CHARGES = { # NOTE : units deliberately omitted here (become applied to entire charge array)
        'H' :  0.417,
        'O' : -0.843
    }

    water = Molecule.from_smiles('O')
    water.name = 'water_TIP3P'
    water.partial_charges = [TIP3P_ATOM_CHARGES[atom.symbol] for atom in water.atoms]*offunit.elementary_charge
    for atom in water.atoms:
        atom.metadata['residue_name'] = 'HOH'

    return water


# predefine water file, if not already present
_water_path = _MODULE_PATH / 'water_TIP3P.sdf'
if not _water_path.exists():
    water = generate_water_TIP3P()
    topology.package_atom_metadata(water, in_place=True)
    topology.save_molecule(_water_path, water)

# register Molcules for all registered solvents
for path in _MODULE_PATH.iterdir():
    if path.suffix == '.sdf':
        offmol = Molecule.from_file(path) # load molecules from file
        globals()[path.stem] = topology.unpackage_atom_metadata(offmol, in_place=False) # recover metadata and register solvent Molecules locally



'''Utilities from describing the various parameters and settings possessed by OpenMM objects'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Union
from openmm import NonbondedForce, System

from ...genutils.attrs import compile_argfree_getable_attrs
from ...genutils.textual.prettyprint import dict_to_indented_str


# REFERENCE TABLES FOR NONBONDED METHODS
NONBOND_METHOD_KEY : str = 'NonbondedMethod'
NONBOND_CUTOFF_METHOD_NAMES = (
    'NoCutoff',
    'CutoffNonPeriodic',
    'CutoffPeriodic',
    'Ewald',
    'PME',
    'LJPME',
)
NONBOND_CUTOFF_METHODS = {
    idx : method_name
        for idx, method_name in sorted( # sort in ascending order by integer code
            (getattr(NonbondedForce, method_name), method_name)
                for method_name in NONBOND_CUTOFF_METHOD_NAMES
        )
}

# DESCRIBING OpenMM PARAMETERS
def describe_forces(ommsys : System, as_str : bool=False) -> Union[str, dict[str, dict[str, Any]]]:
    '''Provides a dictionary (keyed by force names) which summarizes the parameters of each Force in an OpenMM system'''
    force_desc_dict = {}
    for force in ommsys.getForces():
        force_attrs = compile_argfree_getable_attrs(force, getter_re='\Aget', repl_str='') # getter string here asserts that "get" is at the start of the attribute name
        force_attrs['Type'] = type(force).__name__
        
        if (nonbond_id := force_attrs.get(NONBOND_METHOD_KEY)) is not None:
            force_attrs[nonbond_id] = NONBOND_CUTOFF_METHODS.get(nonbond_id, nonbond_id) # attempt a lookup of the nonbonded method name
        force_desc_dict[force.getName()] = force_attrs

    if as_str:
        return dict_to_indented_str(force_desc_dict)
    return force_desc_dict


'''Tools for generating plots and other graphics'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Iterable, Optional
from pathlib import Path
from PIL.Image import Image

import numpy as np
from pandas import DataFrame
from math import ceil

import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, Colormap
from matplotlib.colorbar import Colorbar, ColorbarBase
from mpl_toolkits.axes_grid1 import make_axes_locatable

ColorMapper = Callable[[float], tuple[int, ...]]


# PLOT SIZING FUNCTIONS
def presize_subplots(nrows : int, ncols : int, scale : float=15.0, elongation : float=1.0) -> tuple[plt.Figure, np.ndarray[plt.Axes]]:
    '''
    Prepare a grid of predetermined number of matplotlib subplot axes of a particular size and aspect ratio
    Returns the resulting Figure and array of individual subplot Axes 
    '''
    aspect = (nrows / ncols) * elongation
    return plt.subplots(nrows, ncols, figsize=(scale, aspect*scale))

def plot_df_props(x_dframe : DataFrame, y_dframe : DataFrame, df_label : Optional[str]=None, nrows : int=None, ncols : int=None, **plot_kwargs) -> tuple[plt.Figure, plt.Axes]:
    '''Takes a DataFrame populated with polymer time series property data and generates sequential plots'''
    num_series = len(y_dframe.columns)
    if (nrows is None) and (ncols is None): # TODO : deprecate this terribleness in favor of smart axis sizing to aspect ratio, move to presize_subplots()
        nrows = 1
        ncols = num_series
    if (nrows is None):
        nrows = ceil(num_series / ncols)
    if (ncols is None):
        ncols = ceil(num_series / nrows)
    assert(nrows * ncols) >= num_series

    fig, ax = presize_subplots(nrows=nrows, ncols=ncols, **plot_kwargs)
    if not isinstance(ax, np.ndarray):
        ax = np.array([ax]) # convert singleton subplots into arrays so that they don;t break when attempting to be flattened

    for curr_ax, (name, y_dframe) in zip(ax.flatten(), y_dframe.items()):
        curr_ax.plot(x_dframe, y_dframe, label=df_label)
        curr_ax.set_xlabel(x_dframe.columns[0])
        curr_ax.set_ylabel(name)

    return fig, ax


# GENERAL PLOTTING FUNCTIONS
def scatter_3D(array : np.array) -> None:
    '''Plot an Nx3 array of (x, y, z) coordinate sets in 3-D'''
    x, y, z = array.T

    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')
    ax.scatter(x, y, z)


# COLORBAR AND COLORMAP FUNCTIONS
def make_cmapper(cmap_name : str, vmin : float, vmax : float) -> ColorMapper:
    '''Wrapper for making normalized color map function'''
    cmap = plt.get_cmap(cmap_name)
    norm = Normalize(vmin, vmax)

    def cmapper(val_to_map : float) -> tuple[int, ...]:
        '''Actual colormapping function'''
        return cmap(norm(val_to_map))
    
    return cmapper

def plot_image_with_colorbar(image : Image, cmap : Colormap, norm : Normalize, label : str='', ticks : Optional[list[float]]=None, orient : Optional[str]=None, dim : int=8, aspect : float=1/1) -> tuple[plt.Figure, plt.Axes]:
    '''Plots a PIL image with a colorbar and norm of ones choice'''
    CAX_LOC_OPTIONS = {
        'vertical'   : 'right',
        'horizontal' : 'bottom'
    }

    fig, ax = plt.subplots(figsize=(dim, aspect*dim))
    mpim = ax.imshow(image, cmap=cmap, norm=norm)

    width, height = image.size
    if orient is None:
        orient = 'vertical' if (height > width) else 'horizontal' # if no orientation is provided, place along longest axis
    cax_loc = CAX_LOC_OPTIONS[orient] # will raise KeyError if invalid orientation is provided

    div = make_axes_locatable(ax) # allow for creatioin of separate colorbar axis
    ax.set_axis_off() # prevent ticks from interrupting image
    cax = div.append_axes(cax_loc, size='5%', pad='2%')
    cbar = fig.colorbar(mpim, cax=cax, label=label, ticks=ticks, orientation=orient)

    return fig, ax

def draw_colorbar(cmap_name : str, vmin : float, vmax : float, label : str, save_path : Path=None) -> Colorbar:
    '''Create a matplotlib colorbar object with appropriate norm, color scale, and labels'''
    cmap = plt.get_cmap(cmap_name)
    norm = Normalize(vmin, vmax)
    
    fig = plt.figure()
    ax = fig.add_axes([0.9, 0.1, 0.05, 0.95]) # TODO : generalize this sizing
    ticks = [vmin, 0, vmax]

    cbar = ColorbarBase(ax, orientation='vertical', cmap=cmap, norm=norm, ticks=ticks, label=label)
    
    if save_path is not None:
        fig.savefig(save_path, bbox_inches='tight')
        plt.close()

    return cbar

def label_discrete_cmap(cmap : Colormap, color_names : Iterable[str], hues_per_color : int=1) -> tuple[dict[str, tuple[int, int, int, int]], np.ndarray]:
    '''Generate named labels for discrete colormaps (e.g. Qualitative colormaps)
    Accepts the target Colormap, a collection of the unique colors present (in order), and the number of distinct hues per color
    Returns a dict mapping the named colors, as well as an MxNx4 array of the RGBA colors which can be visualized as an image'''
    n_color_types = len(color_names)
    n_colors = n_color_types * hues_per_color
    if hues_per_color > 1:
        color_names = [f'{color}{i}' for color in color_names for i in range(hues_per_color)]

    sample_points = np.linspace(1 / (2*n_colors), 1, num=n_colors) # start in middle of first bin, sample subsequent middles up to unity
    color_samples = [cmap(sample) for sample in sample_points]

    color_arr = np.array(color_samples).reshape(n_color_types, hues_per_color, 4) # assumes RGBA, as returned by calling matplotlib Colormaps
    color_dict = {
        color_name : rgba
            for color_name, rgba in zip(color_names, color_samples)
    }

    return color_dict, color_arr


'''Tools for determining how many and which bondable ports are in an RDKit Mol'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Generator, Iterable, Optional
from itertools import (
    combinations,
    product as cartesian_product,
)
from rdkit.Chem.rdchem import Mol

from .portlib import Port, get_ports, MolPortError
from ...genutils.iteration import iter_len


def get_total_port_degree_on_atom(rdmol : Mol, atom_id : int, target_flavor : Optional[int]=None):
    '''Get the maximum combined bond degree for all ports of a particular flavor on an atom (or of all flavors, if None is provided)'''
    return sum(
        int(port.bond.GetBondTypeAsDouble()) # sum over all matching-flavor ports on an atoms, weighting by bond order
            for port in get_ports(rdmol, target_atom_id=atom_id, target_flavor=target_flavor)
    )

def max_bondable_order_between_atoms(rdmol : Mol, atom_id_1 : int, atom_id_2 : int, target_flavor : int) -> int: # TODO : update or possible deprecate this
    '''Return the highest possible bond order which could be created between a pair of atoms''' # NOTE : can't just count number of bondable pairs, since these include all possible permutations
    if target_flavor is None:
        raise NotImplementedError # TODO : find better way to handle this (lack of selectivity with no flavor prevents accurate bondability estimation)
    
    return min( # limited by which atom has the fewest bonding sites
        get_total_port_degree_on_atom(rdmol, atom_id_1, target_flavor=target_flavor),
        get_total_port_degree_on_atom(rdmol, atom_id_2, target_flavor=target_flavor),
    )

def enumerate_bondable_port_pairs(ports_1 : Iterable[Port], ports_2 : Iterable[Port]) -> Generator[tuple[Port, Port], None, None]:
    '''Generate all possible pairs of Ports between two molecules which could be bonded together'''
    for port_1, port_2 in cartesian_product(ports_1, ports_2):
        if Port.are_bondable(port_1, port_2):
            yield (port_1, port_2)

def enumerate_bondable_port_pairs_internal(ports : Iterable[Port]) -> Generator[tuple[Port, Port], None, None]: # TODO : find way to merge this with enumerate_bondable_port_pairs
    '''Generate all possible pairs of Ports within a single molecule which could be bonded together'''
    for port_1, port_2 in combinations(ports, 2):
        if Port.are_bondable(port_1, port_2):
            yield (port_1, port_2)

# def _get_bondable_ports_internal(rdmol : Mol) -> Generator[tuple[Port, Port], None, None]:
#     '''Generate all possible pairs of Ports within a single molecule which could be bonded together'''
#     return get_bondable_ports(rdmol, rdmol) # currently, this overcounts by exactly double, as a pair (p1, p2) and its copair (p2, p1) are considered distinct

def get_bondable_port_pairs(rdmol : Mol, atom_id_1 : Optional[int]=None, atom_id_2 : Optional[int]=None, flavor_pair : tuple[Optional[int], Optional[int]]=(None, None)) -> Generator[tuple[Port, Port], None, None]:
    '''Get every pair of ports within an RDKit Mol which could be bonded and match a specified flavor pair (could be partially or fully None for less specificity)
    Can optionally localize search to only have port bridegehead coinciding with atoms at particular indices
    Pairs are returned n descending order of linker atom index to simplify atom removal if modifying later'''
    # possible_port_pairs = enumerate_bondable_port_pairs_internal( # unpack values after iterating; TODO : make this internal to avoid double counting
    #     chain( # concatenate together two generators (one for each atom-flavor pair)
    #         *(get_ports(rdmol, target_atom_id=atom_id, target_flavor=flavor) # NOTE : need star unpacking here to pass ports onto thru chain (rather than just passing the two generators)
    #             for (atom_id, flavor) in zip((atom_id_1, atom_id_2), flavor_pair)
    #         )
    #     )
    # )
    possible_ports = set() # treated as set to account for uniqueness
    for atom_id in (atom_id_1, atom_id_2): # TODO : find less nested way to do this
        for flavor in flavor_pair:
            for port in get_ports(rdmol, target_atom_id=atom_id, target_flavor=flavor):
                possible_ports.add(port) 

    for (port_1, port_2) in enumerate_bondable_port_pairs_internal(possible_ports):
        if Port.are_bondable(port_1, port_2):
            if port_1.linker.GetIdx() >= port_2.linker.GetIdx(): # sort in-place to ensure highest-index linker is first (MATTER FOR ORDER OF ATOM REMOVAL)
                yield (port_1, port_2)
            else:
                yield (port_2, port_1)

def get_first_bondable_port_pair(rdmol : Mol, atom_id_1 : Optional[int]=None, atom_id_2 : Optional[int]=None, flavor_pair : tuple[Optional[int], Optional[int]]=(None, None)) -> tuple[Port, Port]:
    '''Get the first pair of ports between atoms at given indices which are bondable, raising Exception if none exists'''
    try:
        return next(get_bondable_port_pairs(rdmol, atom_id_1=atom_id_1, atom_id_2=atom_id_2, flavor_pair=flavor_pair))
    except StopIteration:
        raise MolPortError(f'No bondable ports with flavors {flavor_pair} exist within Mol')

def get_num_bondable_port_pairs(rdmol : Mol, atom_id_1 : Optional[int]=None, atom_id_2 : Optional[int]=None, flavor_pair : tuple[Optional[int], Optional[int]]=(None, None)) -> int:
    '''Count how many bondable port pairs exist within a Mol'''
    return iter_len(get_bondable_port_pairs(rdmol, atom_id_1=atom_id_1, atom_id_2=atom_id_2, flavor_pair=flavor_pair))


'''Representations and computation methods for continued fractions and ration approximations to real numbers'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union, Generator, Iterable, Type, TypeAlias, TypeVar
from ...genutils.typetools.numpytypes import Shape, NPInt

import numpy as np

Real : TypeAlias = Union[float, int]
I = TypeVar('I', bound=Union[int, NPInt])


# CONSTANT DEFAULT PARAMETERS SHARED AMONGST MANY FUNCTIONS BELOW
DEFAULT_INT_TYPE : Type = np.int64
DEFAULT_EPS = 1E-8
DEFAULT_TOL = 1E-6

# CONTINUED FRACTIONS AND PARTIAL CONVERGENTS ("CONTINUANTS")
def continuant_matrix(a : int, int_type : I=DEFAULT_INT_TYPE) -> np.ndarray[Shape[2, 2], int]:
    '''Homographic matrix for a evaluating the next continuant from a continued fraction coefficient'''
    return np.array([
        [a, 1],
        [1, 0]
    ], dtype=int_type)

def _continuant_matrices_from_coeffs(coeffs : Iterable[int], int_type : I=DEFAULT_INT_TYPE) -> Generator[np.ndarray[Shape[2, 2], I], None, None]:
    '''Generate a sequence of continuant matrices representing adjacent steps in unravelling Euclidean division'''
    M = np.eye(2, dtype=int_type) # initialize first two pairs as [1, 0] and [0, 1]
    for c in coeffs:
        M = M @ continuant_matrix(c)
        yield M

def continued_fraction_to_continuants(coeffs : Iterable[int], int_type : I=DEFAULT_INT_TYPE) -> Generator[tuple[int, int], None, None]:
    '''Fold a sequence of continued fraction coefficients into successive continuants (rational approximations)'''
    for M in _continuant_matrices_from_coeffs(coeffs, int_type=int_type):
        yield M[:, 0]

# EUCLIDEAN ALGORITHM VARIANTS
def real_to_continued_fraction_coeffs(x : Real, eps : float=DEFAULT_EPS, int_type : I=DEFAULT_INT_TYPE) -> Generator[int, None, None]:
    '''Euclidean algorithm on a real number to produce the integral continued fraction coefficients'''
    while True:
        n, rem = divmod(x, 1)
        if isinstance(n, float):
            assert(n.is_integer())
        # yield int(n)
        yield int_type(n)
        
        if abs(rem - 0) < eps:
            break # TOSELF : can't make part of while condition (without redundant pre-code) due to initial Euclidean division step
        x = 1 / rem

def extended_euclidean_algorithm(a : int, b : int, int_type : I=DEFAULT_INT_TYPE) -> tuple[int, int, int]:
    '''Compute the Extended Euclidean Algorithm between two integers "a" and "b"
    Returns the greatest common divisor of a and b, along with a pair of Bezout coefficients"x" and "y" which satisfy a*x + b*y = gcd(a, b)'''
    quotients : list[int] = []
    ai, bi = a, b # make copies to avoid mutation and to have access to original values after iteration
    while bi:
        quot, rem = divmod(ai, bi)
        ai, bi = bi, rem
        quotients.append(-quot) # NOTE: negative sign is intentional here

    _gcd = ai # NOTE: underline is to avoid name clash with builtin math.gcd
    *_, bezout_coeffs = _continuant_matrices_from_coeffs(quotients, int_type=int_type) # NOTE: do have access to original a, b values in first column, but signs are annoying to keep track of
    y, x = bezout_coeffs[:, -1]
    assert((a*x + b*y) == _gcd) # verify that Bezout's identity is actually satisfied

    return _gcd, x, y
bezout_coeffs = euclidean_extended = extended_euclidean_algorithm

# RATIONAL APPROXIMATIONS FROM CONTINUED FRACTIONS
def rational_approxes(x : Real, tol : float=DEFAULT_TOL, eps : float=DEFAULT_EPS) -> Generator[tuple[int, int], None, None]:
    '''Unfold a real number into its continued fraction representation, then generate successive continuants of it'''
    for p, q in continued_fraction_to_continuants(real_to_continued_fraction_coeffs(x, eps=eps)):
        yield p, q
        if abs(p/q - x) < tol:
            break

def best_rational_approx(x : Real, tol : float=DEFAULT_TOL, eps : float=DEFAULT_EPS) -> tuple[int, int]:
    '''Provide a rational approximation to a value with the smallest denominator that is within some tolerance'''
    *_, best = rational_approxes(x, tol=tol, eps=eps)
    return best



'''For converting macroscopic parameters (such as concentration, bulk density, etc) into microscopic parameters for simulations'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Union

from math import ceil
from rdkit.Chem import Descriptors, Mol

from openmm.unit import gram, centimeter, mole
from openmm.unit import Quantity, AVOGADRO_CONSTANT_NA

from openff.toolkit import Molecule, Topology
from openff.units import Quantity as OFFQuantity

from ....unitutils.dimensions import is_volume
from ..unitsys import allow_openff_units, openff_to_openmm


# MASS
def molecular_weight(rdmol : Mol, exact_isotopes : bool=False) -> Quantity:
    '''Compute the molecular weight of an RDKit Mol'''
    MW = Descriptors.ExactMolWt(rdmol) if exact_isotopes else Descriptors.MolWt(rdmol)
    return MW * (gram / mole) # attach units

def offmol_mass(offmol : Molecule, as_openmm : bool=False) -> Union[Quantity, OFFQuantity]:
    '''Compute the molecular mass of an OpenFF Molecule (in Daltons)
    Returns mass with OpenFF-style units, or with OpenMM-style units if as_openmm=True'''
    mass = sum(atom.mass for atom in offmol.atoms)
    if as_openmm:
        return openff_to_openmm(mass)
    return mass

def offtop_mass(offtop : Topology, as_openmm : bool=False) -> Union[Quantity, OFFQuantity]:
    '''Compute the molecular mass of an OpenFF Topology (in Daltons)
    Returns mass with OpenFF-style units, or with OpenMM-style units if as_openmm=True'''
    mass = sum(offmol_mass(offmol, as_openmm=False) for offmol in offtop.molecules)
    if as_openmm:
        return openff_to_openmm(mass) # perform conversion at end, to avoid multiple intermediate conversions and avoid initializing sum() with Quantity
    return mass

# DENSITY
@allow_openff_units
def number_density(density : Quantity, MW : Quantity) -> Quantity:
    '''
    Determine the number of solvent molecules per unit volume from known physical constants
    For best results, provide arguments as Quantities with associated units
    '''
    return (density / MW) * AVOGADRO_CONSTANT_NA

# NUMBER
@allow_openff_units
def num_mols_in_box(mol : Union[Mol, Molecule, Topology], box_vol : Quantity, density : Quantity) -> int:
    '''Return the number of particles/molecules needed to fill a box of given volume to the specified density'''
    assert(is_volume(box_vol.unit))

    if isinstance(mol, Mol):
        MW = molecular_weight(mol, exact_isotopes=False)
    elif isinstance(mol, Molecule):
        MW = offmol_mass(mol, as_openmm=True)
    elif isinstance(mol, Topology):
        MW = offtop_mass(mol, as_openmm=True)
    else:
        raise TypeError(f'No molecular weight calculations implemented for object of type "{type(mol).__name__}"')

    return ceil(box_vol * number_density(density=density, MW=MW))


'''For generating periodically-tiled topologies from RDKit Mols'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import numpy as np

from rdkit import Chem
from rdkit.Chem import rdGeometry, rdMolTransforms, Mol

from ...genutils.typetools.numpytypes import Shape, N
from ...maths.linearalg import affine
from ...maths.lattices.coordinates import Coordinates


def rdmol_effective_radius(rdmol : Chem.Mol, conf_id : int=0) -> float:
    '''Determine an effective radius of influence of a molecule such that all atoms are within this radius'''
    conformer = rdmol.GetConformer(conf_id)
    positions = Coordinates(conformer.GetPositions())
    dists_to_centroid = positions.dists_to_centroid()

    # polymerist-agnostic implementation
    # centroid  = rdMolTransforms.ComputeCentroid(conformer, ignoreHs=True) # = positions.mean(axis=0) # TOSELF : mean positiona dn centroid give slightly different result, but don't affect validity of tiling
    # dists_to_centroid = np.linalg.norm(positions - centroid, axis=1)

    return dists_to_centroid.max()

def tile_lattice_with_rdmol(rdmol : Mol, lattice_points : np.ndarray[Shape[N, 3], float], rotate_randomly : bool=True, conf_id : int=0) -> Mol:
    '''
    Generate a tiled topology of copies of an RDKit molecule, transformed to occupy the same relative positions as points on the given lattice with unit dimensions
    if "random_rotations" is True, each occurrence of the molecule will also have a random rotation applied to it. Mol is NOT modified at any point in the procedure
    '''
    if rdmol.GetNumConformers() < 1:
        raise ValueError('Molecule must have at least one conformer to be tiled')
    
    conformer = rdmol.GetConformer(conf_id)
    centering = rdMolTransforms.ComputeCanonicalTransform(conformer, ignoreHs=True) # translation which centers the given conformer

    tiled_topology = None
    for point in lattice_points:
        translation = affine.xyzTrans(*point) # translation to the current point on the lattice
        rotation    = affine.randRot(about_x=rotate_randomly, about_y=rotate_randomly, about_z=rotate_randomly) # rotation about either all axes or the identity matrix
        transform = translation @ rotation @ centering # NOTE : (right-to-left) order matters here!! Must first move to origin, then rotate, and then translate to latice point 

        clone = Chem.Mol(rdmol) # make copy to avoid modifying ariginal
        rdMolTransforms.TransformConformer(clone.GetConformer(0), transform) # apply transform to 
        
        if tiled_topology is None:
            tiled_topology = clone
        else:
            tiled_topology = Chem.CombineMols(tiled_topology, clone)

    return tiled_topology


'''Classes for representing information about reaction mechanisms and tracing bonds and atoms along a reaction'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import ClassVar, Iterable, Optional, Sequence, Union
from dataclasses import dataclass, field

import re
from io import StringIO
from pathlib import Path
from functools import cached_property

from rdkit.Chem import rdChemReactions, Mol

from .reactexc import BadNumberReactants
from ..bonding._bonding import combined_rdmol
from ..labeling.molwise import ordered_map_nums
from ..labeling.bondwise import get_bonded_pairs_by_map_nums

from ...genutils.decorators.functional import allow_string_paths, allow_pathlib_paths
from ...genutils.sequences.discernment import DISCERNMENTSolver
from ...smileslib.substructures import num_substruct_queries_distinct


# REACTION INFORMATICS CLASSES
@dataclass
class RxnProductInfo:
    '''For storing atom map numbers associated with product atoms and bonds participating in a reaction'''
    prod_num : int
    reactive_atom_map_nums : list[int] = field(default_factory=list)

    new_bond_ids_to_map_nums : dict[int, tuple[int, int]] = field(default_factory=dict)
    mod_bond_ids_to_map_nums : dict[int, tuple[int, int]] = field(default_factory=dict)
    
RXNNAME_RE = re.compile(r'^\t*(?P<rxnname>.*?)\n$')
class AnnotatedReaction(rdChemReactions.ChemicalReaction):
    '''
    RDKit ChemicalReaction subclass with additional useful information about product atom and bond mappings and reaction naming
    Initialization must be done either via AnnotatedReaction.from_smarts, AnnotatedReaction.from_rdmols, or AnnotatedReaction.from_rxnfile
    '''
    # line number in .rxn file where (optional) name of reaction should be located (per the CTFile spec https://discover.3ds.com/sites/default/files/2020-08/biovia_ctfileformats_2020.pdf) 
    _RXNNAME_LINE_NO : ClassVar[int] = 1
    
    # LOADING/EXPORT METHODS
    @classmethod
    def from_smarts(cls, rxn_smarts : str) -> 'AnnotatedReaction':
        '''Iinstantiate reaction from mapped SMARTS string'''
        return cls(rdChemReactions.ReactionFromSmarts(rxn_smarts))
    
    # NOTE : cannot analogous implement "from_smiles" classmethod, as rdChemreactions does not support initialization from SMILES (only SMARTS)

    def to_smarts(self) -> str:
        '''Export reaction as mapped SMARTS string'''
        return rdChemReactions.ReactionToSmarts(self) # TODO : implement * -> R replacement here (rather than in rxn file I/O)

    @property
    def smarts(self) -> str:
        '''Mapped SMARTS string representation of reaction'''
        return self.to_smarts()

    def to_smiles(self) -> str:
        '''Export reaction as mapped SMILES string'''
        return rdChemReactions.ReactionToSmiles(self) # TODO : implement * -> R replacement here (rather than in rxn file I/O)

    @property
    def smiles(self) -> str:
        '''Mapped SMILES string representation of reaction'''
        return self.to_smiles()

    @classmethod
    def from_rdmols(cls, reactant_templates : Iterable[Mol], product_templates : Iterable[Mol], agent_templates : Optional[Iterable[Mol]]=None) -> 'AnnotatedReaction':
        '''For instantiating reactions directly from molecules instead of SMARTS strings'''
        # label atoms as belonging to reactant or product via RDKit 'magic' internal property (1 = reactant, 2 = product, 3 = agent)
        if agent_templates is None:
            agent_templates = []

        template_role_map = { # RDKit "magic" aliases which define what role each atom plays in a reaction
            1 : reactant_templates,
            2 : product_templates,
            3 : agent_templates,
        }

        for mol_rxn_role, templates in template_role_map.items():
            for template in templates: # TODO : implement non-in-place assignment of these properties
                for atom in template.GetAtoms():
                    atom.SetIntProp('molRxnRole', mol_rxn_role) 
        rxn_mol = combined_rdmol(*reactant_templates, *product_templates, *agent_templates, assign_map_nums=False, editable=False) # kwargs are explicitly needed here

        return cls(rdChemReactions.ReactionFromMolecule(rxn_mol))

    # I/O METHODS
    @classmethod
    @allow_string_paths
    def rxnname_from_rxnfile(cls, rxnfile_path : Union[str, Path]) -> str:
        '''Extract the reaction name from a (properly-formatted) .RXN file'''
        with rxnfile_path.open('r') as file:
            for i, line in enumerate(file):
                if i == cls._RXNNAME_LINE_NO:
                    return re.match(RXNNAME_RE, line).group('rxnname')
                
    @property
    def rxnname(self) -> str:
        '''A string handle associated with this reaction'''
        return getattr(self, '_rxnname', '') # default to the empty string if unset
    
    @rxnname.setter
    def rxnname(self, new_rxnname : str) -> None:
        '''Set a new name for the current reaction'''
        self._rxnname = new_rxnname

    @classmethod
    @allow_pathlib_paths
    def from_rxnfile(cls, rxnfile_path : Union[str, Path]) -> 'AnnotatedReaction':
        '''For instantiating reactions directly from MDL .rxn files'''
        rxn = cls(rdChemReactions.ReactionFromRxnFile(rxnfile_path))
        rxn.rxnname = cls.rxnname_from_rxnfile(rxnfile_path)

        return rxn
    
    @allow_string_paths
    def to_rxnfile(self, rxnfile_path : Union[str, Path], wilds_to_R_groups : bool=True) -> None:
        '''Save reaction to an MDL .RXN file. Replaces ports with R-groups to enable proper loading'''
        rxn_block = rdChemReactions.ReactionToRxnBlock(self)
        if wilds_to_R_groups:
            rxn_block = rxn_block.replace('*', 'R')

        with rxnfile_path.open('w') as rxnfile:
            for i, line in enumerate(StringIO(rxn_block)):
                if (i == self._RXNNAME_LINE_NO):
                    rxnfile.write(f'\t\t\t{self.rxnname}\n')
                else:
                    rxnfile.write(line)

    # RXN INFO METHODS
    @cached_property
    def reacting_atom_map_nums(self) -> list[int]:
        '''List of the map numbers of all reactant atoms which participate in the reaction'''
        return [
            self.GetReactantTemplate(reactant_id).GetAtomWithIdx(atom_id).GetAtomMapNum()
                for reactant_id, reacting_atom_ids in enumerate(self.GetReactingAtoms())
                    for atom_id in reacting_atom_ids
        ]
    
    @cached_property
    def map_nums_to_reactant_nums(self) -> dict[int, int]:
        '''Back-map yielding the index of the source reactant for the atom of each map number'''
        return {
            atom.GetAtomMapNum() : i
                for i, react_template in enumerate(self.GetReactants())
                    for atom in react_template.GetAtoms()
        }
    
    @cached_property
    def product_info_maps(self) -> dict[int, RxnProductInfo]:
        '''Map from product index to information about reactive atoms and bonds in that product'''
        # map reacting atoms and bonds for each product
        prod_info_map = {}
        for i, product_template in enumerate(self.GetProducts()):
            prod_info = RxnProductInfo(i)
            prod_info.reactive_atom_map_nums = [
                map_num
                    for map_num in self.reacting_atom_map_nums
                        if map_num in ordered_map_nums(product_template)
            ]

            for bond_id, atom_id_pair in get_bonded_pairs_by_map_nums(product_template, *prod_info.reactive_atom_map_nums).items(): # consider each pair of reactive atoms
                map_num_1, map_num_2 = map_num_pair = tuple(
                    product_template.GetAtomWithIdx(atom_id).GetAtomMapNum()
                        for atom_id in atom_id_pair
                )
                
                if self.map_nums_to_reactant_nums[map_num_1] == self.map_nums_to_reactant_nums[map_num_2]: # if reactant IDs across bond match, the bond must have been modified (i.e. both from single reactant...)
                    prod_info.mod_bond_ids_to_map_nums[bond_id] = map_num_pair
                else: # otherwise, bond must be newly formed (spans between previously disjoint monomers) 
                    prod_info.new_bond_ids_to_map_nums[bond_id] = map_num_pair
            prod_info_map[i] = prod_info
        
        return prod_info_map
    
    def valid_reactant_ordering(self, reactants : Sequence[Mol], as_mols : bool=True) -> Optional[list[Mol]]:
        '''
        Given an RDKit chemical reaction mechanism and a sequence of reactant Mols, will determine if there is
        an ordering of the reactants which is compatible with the reactant templates defined in the reaction

        Returns the first found ordering, or NoneType if no such ordering exists
        Ordering returned as list of Chem.Mol objects if as_mols == True, or as list of ints otherwise
        '''
        # 0) Preliminary quick check on number of reactants; can discount a bad reactant collection prior to more expensive check
        num_reactants_provided = len(reactants)
        num_reactant_templates_in_mechanism = self.GetNumReactantTemplates()

        if num_reactants_provided != num_reactant_templates_in_mechanism:
            raise BadNumberReactants(f'{self.__class__.__name__} expected {num_reactant_templates_in_mechanism} reactants, but {num_reactants_provided} were provided')

        # if number of fragments is correct, perform more complex of whether a molecule-unique substructure selection exists
        ## 1) Pre-compilation of template substructure indices
        reactant_template_indices : list[int] = [templ_idx for templ_idx in range(num_reactant_templates_in_mechanism)]
        reactant_templates_by_index : dict[int, Mol] = {
            templ_idx : self.GetReactantTemplate(templ_idx)
                for templ_idx in reactant_template_indices
        }

        template_substruct_ids_by_reactant_ids : dict[int, list[int]] = { 
            react_idx : [ # associate with each reactant a substructure "word"...
                templ_idx # ... consisting of the indices of reactant template indices present as substructure(s) in the reactant...
                    for templ_idx, react_templ in reactant_templates_by_index.items()
                        for _ in range(num_substruct_queries_distinct(reactant, react_templ)) # ...which appear as many times as the substructure is uniquely present
            ]
                for react_idx, reactant in enumerate(reactants)
        } # TOSELF: implemented this w/ indices (rather than directly using Mols, which are hashable) because the same mol might return distinct hashes (cannot evaluate self-equality via __eq__ directly)

        ## 2) Choice-finding to see if a selection of reactants matching the templates exists
        reactant_ordering_planner = DISCERNMENTSolver(template_substruct_ids_by_reactant_ids)

        if reactant_ordering_planner.solution_exists(reactant_template_indices, unique_bins=True):
            reactant_ordering : tuple[int] = next(reactant_ordering_planner.enumerate_choices(
                word=reactant_template_indices,
                unique_bins=True) # need to have unique bins so a single reactant is not taken more than once
            )
            if as_mols:
                return [reactants[i] for i in reactant_ordering] # get first valid ordering
            return reactant_ordering
        else: # NOTE: this else clause is not strictly necessary (would otherwise return None anyway), but prefer to have it for explicitness
            return None 


'''Tools for converting back and forth between integers and ordinal number words and prefices (https://en.wikipedia.org/wiki/Ordinal_numeral)'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

def ordinal_suffix_from_int(n : int) -> str:
    '''Produce the appropriate word suffix for an integer in sequential order
    E.g 1 -> "st" as in "first", 17 -> "th" as in "seventeenth, etc.'''
    n = int(abs(n)) # will tolerate negative values and floats which look like ints
    dsuff = {
        1 : 'st',
        2 : 'nd',
        3 : 'rd'
    }
    if (1 <= (i := n % 10) <= 3) and not (11 <= (n % 100) <= 13): # check for special case of 1, 2, and 3 EXCEPT when occuring in 11, 12, or, 13
        return dsuff[i]
    return 'th' # everything else ends in "th"

def ordinal_suffix_from_int_alt(n : int) -> str:
    '''Produce the appropriate word suffix for an integer in sequential order
    E.g 1 -> "st" as in "first", 17 -> "th" as in "seventeenth, etc.
    
    An alternative, slightly-slower but aesthetically-pleasing implementation''' # taken from SO https://stackoverflow.com/questions/9647202/ordinal-numbers-replacement/ 
    n = int(abs(n)) # will tolerate negative values and floats which look like ints
    if (n % 100) in (11, 12, 13): # annoying special cases for first few tens values:
        return 'th'
    return ['th', 'st', 'nd', 'rd', 'th'][min(n % 10, 4)] # min and extra "th" at end give sneaky way of defaulting to 'th'

def ordinal_from_int(n : int) -> str:
    '''Produce the word representation of an integer sequential order
    E.g 1 -> "1st", 17 -> "seventeenth", 33 -> "33rd", etc.'''
    return str(n) + ordinal_suffix_from_int(n)


'''For representing monomer information'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

import logging
LOGGER = logging.getLogger(__name__)

from typing import Generator, Optional, Iterable, Union
from dataclasses import dataclass, field

from itertools import cycle
from collections import defaultdict

from rdkit import Chem

from ...genutils.iteration import iter_len
from ...genutils.fileutils.jsonio.jsonify import make_jsonifiable

from ...smileslib.primitives import Smarts, is_valid_SMARTS
from ...rdutils.bonding.portlib import get_num_ports


# MAIN REPRESENTATION CLASS
@make_jsonifiable
@dataclass
class MonomerGroup:
    '''Stores collections of residue-labelled monomer SMARTS'''
    monomers : dict[str, Union[Smarts, list[Smarts]]] = field(default_factory=dict)
    term_orient : dict[str, str] = field(default_factory=dict) # keys are either "head" or "tail", values are the names of residues in "monomers"

    # MONOMER ADDITION AND VALIDATION
    def __post_init__(self) -> None:
        # Encase bare SMARTS into lists and check that all monomer SMARTS are valid
        monomers_init = self.monomers # store inputted values
        self.monomers = {} # clear monomers and re-add one-at-a-time
        for resname, smarts in monomers_init.items():
            self.add_monomer(resname, smarts)
        # DEV: opted to forgo term_orient check for now, as modifying this violates the read-only data model aimed for here
                
    def _add_monomer(self, resname : str, smarts : Smarts) -> None:
        '''Add a new monomer to the templates already stored within, subject to validation checks'''
        if not isinstance(smarts, str): 
            raise TypeError(f'Values of monomers must be either SMARTS strings or lists of SMARTS strings, not "{type(smarts).__name__}"')
        # DEV: include check for empty string? (technically still a valid SMARTS string, but a pretty pathological one at that)
        if not is_valid_SMARTS(smarts):
            raise ValueError(f'Provided invalid monomer SMARTS string for {resname}: "{smarts}"') 
        # DEV: decide whether or not SMILES expansion and spec-compliance should be enforced here or shunted off to the user 
        
        if resname in self.monomers:
            existing_resgroup = self.monomers[resname]
            if isinstance(existing_resgroup, list) and (smarts not in existing_resgroup):
                LOGGER.info(f'Extending existing residue category "{resname}" with SMARTS {smarts}')
                self.monomers[resname].append(smarts)
        else:
            LOGGER.info(f'Creating new residue category "{resname}", containing singular SMARTS ["{smarts}"])')
            self.monomers[resname] = [smarts]
            
    def _add_monomers(self, resname : str, smarts_container : Iterable[Smarts]) -> None:
        '''Add new monomers to the templates already stored within, subject to validation checks, from an iterable container'''
        for smarts in smarts_container:
            self._add_monomer(resname, smarts)
    
    def add_monomer(self, resname : str, smarts : Union[Smarts, Iterable[Smarts]]) -> None:
        '''Register new monomers, either directly from SMARTS or from a container of SMARTS'''
        if isinstance(smarts, Iterable) and not isinstance(smarts, str): # don;t want to insert one character at a time if a string is in fact provided
            self._add_monomers(resname, smarts)
        else:
            self._add_monomer(resname, smarts) # assume any other inputs are singular values or strings 
    
    # DUNDER "MAGIC" METHODS
    def __getitem__(self, resname : str) -> str:
        '''Convenience method to access .monomers directly from instance'''
        return self.monomers[resname] # NOTE: deliberately avoid "get()" here to propagate KeyError
        # BUG: user can directly append to the returned value to forgo monomer validation checks;
        # this is not unit to __getitem__ but rather a consequence of thinly-wrapping builtin types

    def __setitem__(self, resname : str, smarts : Smarts) -> str:
        '''Convenience method to access .monomers directly from instance'''
        self.add_monomer(resname, smarts)
        
    def __hash__(self) -> int:
        '''Hash based on monomer SMARTS and terminal orientation in a canonical order'''
        # TOSELF: this is far from bulletproof, viz. canonicalzation of SMARTS, list value sorting, etc
        return hash(f'{sorted(self.monomers.items())}{sorted(self.term_orient.items())}')
    
    # ATTRIBUTE PROPERTIES AND ALIASES
    @staticmethod
    def is_terminal(monomer : Chem.Mol) -> bool:
        '''Determine whether or not a monomer is terminal'''
        return get_num_ports(monomer) == 1
    
    @property
    def SMARTS(self) -> dict[str, list[Smarts]]:
        '''Alias of legacy "monomers" attribute'''
        return self.monomers # alias of legacy name for convenience
    
    # ITERATION OVER STORED MOLECULE FRAGMENTS
    def iter_rdmols(self, term_only : Optional[bool]=None) -> Generator[tuple[str, Chem.Mol], None, None]:
        '''
        Generate (residue name, RDKit Mol) pairs of all monomers present
        Simplifies iteration over internal lists of monomer Mols

        Can optionally filter by monomer termination:
            term_only=True  -> only terminal monomers
            term_only=False -> only middle monomers
            term_only=None  -> all monomers
        '''
        for resname, SMARTS_list in self.monomers.items():
            for SMARTS in SMARTS_list:
                monomer = Chem.MolFromSmarts(SMARTS)
                if (term_only is None) or (MonomerGroup.is_terminal(monomer) == term_only):
                    yield (resname, monomer)

    def rdmols(self, term_only : Optional[bool]=None) -> dict[str, list[Chem.Mol]]:
        '''
        Returns dict of RDKit Mol lists keyed by residue name

        Can optionally filter by monomer termination:
            term_only=True  -> only terminal monomers
            term_only=False -> only middle monomers
            term_only=None  -> all monomers
        '''
        rdmol_dict = defaultdict(list)
        for resname, rdmol in self.iter_rdmols(term_only=term_only):
            rdmol_dict[resname].append(rdmol)

        return rdmol_dict
    
    def contributions(self, term_only : Optional[bool]=None) -> dict[str, list[int]]:
        '''Returns dict of the number of real (i.e. non-linker) atoms in each residue list'''
        return {
            resname : [mol.GetNumAtoms() - get_num_ports(mol) for mol in mol_list]
                for resname, mol_list in self.rdmols(term_only=term_only).items()
        }
    
    @property
    def n_monomers(self) -> int:
        '''Returns number of present monomers; multiple monomers under the same residue name are considered distinct'''
        return iter_len(self.iter_rdmols(term_only=None))
    
    # END GROUP DETERMINATION 
    def linear_end_groups(self) -> dict[str, tuple[str, Chem.Mol]]:
        '''
        Returns head-and-tail end group residue names and Mol objects as defined by term_orient
        
        If term orient is undefined, will automatically take then first 
        <= 2 terminal groups available to be the end groups
        
        Returns
        -------
        end_groups : dict[str, tuple[str, Chem.Mol]]
            A dict whose keys are any of {'head', 'tail'} and whose
            values are 2-tuples of residue names and Mols for the corresponding monomer
        '''
        if self.term_orient and set(self.term_orient.keys()) == {'head', 'tail'}:
            LOGGER.info(f'Using user-defined terminal group orientation {self.term_orient}')
            monomer_iters = {
                resname : cycle(smarts_list) 
                    for resname, smarts_list in self.rdmols(term_only=True).items()
            } # cycle handles degenerate end group case correctly
            
            return {
                head_or_tail : (resname, next(monomer_iters[resname])) # will raise KeyError if any of the resnames are not present
                    for head_or_tail, resname in self.term_orient.items()
            }
        else:
            term_orient_auto : dict[str, Smarts] = {}
            end_groups_auto  : dict[str, Chem.Mol] = {}
            for head_or_tail, (resname, rdmol) in zip(['head', 'tail'], self.iter_rdmols(term_only=True)): # zip will bottom out early if fewer than 2 terminal monomers are present
                term_orient_auto[head_or_tail] = resname # populate purely for logging
                end_groups_auto[head_or_tail]  = (resname, rdmol)
            LOGGER.warning(f'No valid terminal monomer orientations defined, auto-assigned orientations "{term_orient_auto}"; USER SHOULD VERIFY THIS YIELDS A CHEMICALLY-VALID POLYMER!')
                
            return end_groups_auto
    
    # COMPOSITION METHODS
    def __add__(self, other : 'MonomerGroup') -> 'MonomerGroup':
        '''Content-aware method of merging multiple sets of monomer info via the addition operator'''
        cls = self.__class__
        if not isinstance(other, cls):
            raise NotImplementedError(f'Can only merge {cls.__name__} with another {cls.__name__}, not object of type {type(other)}')
        # TODO: figure out how to handle combination of term group orientation gracefully (ignoring for now)
        return MonomerGroup(monomers={**self.monomers, **other.monomers})

    __radd__ = __add__ # support reverse addition

    # CHEMICAL INFORMATION
    def unique(self, cap_group : Union[Smarts, Chem.Mol]=Chem.MolFromSmarts('[H]-[*]')) -> 'MonomerGroup':
        '''Return a MonomerGroup containing only the unique monomers present, given a particular port saturating group (by default just a hydrogen)'''
        raise NotImplementedError
        # unique_mono = set()
        # for SMARTS in monomer_smarts.values():
        #     monomer = Chem.MolFromSmarts(SMARTS)
        #     clear_atom_map_nums(monomer, in_place=True) 
        #     hydrogenate_monomer_ports(monomer, in_place=True) 
        #     unique_mono.add(Chem.MolToSmiles(monomer)) # TODO : eventually make this SMARTS-based (or even better RDKit Mol-based); can't for now since hydrogenated fragments don't equate

        # return unique_mono

    def is_homopolymer(self) -> bool:
        '''Identify if a polymer is a homopolymer (i.e. only 1 type of middle monomer)'''
        # n_mid, n_term = count_middle_and_term_mono(monomer_smarts) # TODO : reimplement with comparison of port-hydrogenated monomers
        # return (n_mid == 1)
        return (len(self.unique()) == 1) # by definition, a homopolymer only has 1 unique class of monomer

    # GRAPH INFORMATION
    @property
    def is_branchable(self) -> bool:
        '''Whether it is possible to generate a branched polymer from this set of monomers'''
        return any(
            get_num_ports(monomer) > 2
                for (resname, monomer) in self.iter_rdmols(term_only=None)
        )
    
    @property
    def is_linear(self) -> bool:
        '''Whether a group of monomers can ONLY be assembled into a linear chain'''
        return not self.is_branchable

    @property
    def is_linear_homopolymer(self) -> bool:
        '''Identify if a polymer is a linear homopolymer'''
        return self.is_linear and self.is_homopolymer

    @property
    def num_mid_and_term(self) -> tuple[int, int]:
        '''Counts of how many of the monomers are middle vs terminal, respectively'''
        group_counts = [0, 0]
        for (resname, monomer) in self.iter_rdmols(term_only=None): # TODO : consider reimplementing using new term group filtering option
            group_counts[self.is_terminal(monomer)] += 1 # index by bool
        
        return tuple(group_counts) # convert to tuple


'''Tools for simplifying the construction of reaction templates'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass, field
from typing import Iterable, Optional, Union

from rdkit import Chem
from rdkit.Chem.rdchem import Mol

from .reactions import AnnotatedReaction
from ..labeling import molwise
from ..bonding._bonding import combined_rdmol
from ..bonding.permutation import swap_bonds
from ...polymers.monomers import specification


@dataclass # TODO : make JSONifiable
class ReactionAssembler:
    '''Class for producing reaction templates from reactants, bond derangements'''
    reactive_groups  : Iterable[Mol]
    byproducts       : Iterable[Mol]              = field(default_factory=list)
    bond_derangement : dict[int, tuple[int, int]] = field(default_factory=dict)
    rxn_name         : str = ''
    # TODO : incorporate stereo

    @property
    def reactants(self) -> Mol:
        '''Combine cached reactive groups into single, contiguously-numbered Mol for manipulation'''
        # 1) extracting and labelling reactants
        reactants = molwise.assign_contiguous_atom_map_nums(*self.reactive_groups, in_place=False) # needed up-front to display reactants for derangement determination
        return Chem.CombineMols(*reactants) 

    def products(self, show_steps : bool=False) -> Mol:
        '''Generate the product template defined by the provided reactants and bond derangement'''
        if not self.bond_derangement:
            raise ValueError('Must provide non-empty bond derangement')

        # 2) defining and swapping bonds to form product
        products = swap_bonds(Chem.RWMol(self.reactants), self.bond_derangement, show_steps=show_steps) # create editable Mol
        Chem.SanitizeMol(products, sanitizeOps=specification.SANITIZE_AS_KEKULE)

        return products
    
    def products_by_importance(self, combined : bool=True, show_steps : bool=False) -> tuple[Union[Optional[Mol], list[Mol]], Union[Optional[Mol], list[Mol]]]:
        '''Partition reaction products into major and minor/byproducts, each returned as a single Combined Mol'''
        product_partition = main_products, byproducts = [], []
        for product in Chem.GetMolFrags(self.products(show_steps=show_steps), asMols=True):
            for side_query_mol in self.byproducts:
                if product.HasSubstructMatch(side_query_mol) and (product.GetNumAtoms() == side_query_mol.GetNumAtoms()):
                    byproducts.append(product)
                    break # assumes uniquely-provided byproducts (a reasonable assumption)
            else:
                main_products.append(product)

        if not combined:
            return product_partition
        
        return [ # implicit else
            combined_rdmol(*mol_list, assign_map_nums=False, editable=False) if mol_list else None
                for mol_list in product_partition
        ]
    
    @property # TODO : add assertion that reactant and product map num sets match
    def byproduct_map_nums(self) -> tuple[set, set]:
        '''Partitions map numbers present in the product(s) by whether or not they belong to a collection of side products
        Returns a set of map numbers NOT in a side product and set set which are'''
        return [
            set(molwise.ordered_map_nums(product_mol)) if (product_mol is not None) else set()
                for product_mol in self.products_by_importance(combined=True, show_steps=False) # CRITICAL that mols be combined here
        ]

    @property
    def byproduct_relabeling(self) -> dict[int, int]:
        '''Determine a relabeling of the N map numbers in the reaction template which are not any of the given side products
        Relabeling is in standard form, i.e. new labels are taken from the first N natural numbers with the order of labels preserved'''
        map_nums_to_keep, map_nums_to_clear = self.byproduct_map_nums
        
        relabeling = {}
        for i, map_num in enumerate(map_nums_to_keep):
            relabeling[map_num] = (i + 1) # generate standard labeling of preserved map nums

        for map_num in map_nums_to_clear:
            relabeling[map_num] = 0 # unset labels on the unkept atoms

        return relabeling

    def assemble_rxn(self, show_steps : bool=False) -> AnnotatedReaction:
        '''Assemble MDL rxn template from information stored in self'''
        reactants = self.reactants
        products, byproducts = self.products_by_importance(combined=True, show_steps=show_steps)

        if byproducts is not None:
            relabeling = self.byproduct_relabeling
            reactants = molwise.relabel_map_nums(reactants, relabeling=relabeling, in_place=False)
            products  = molwise.relabel_map_nums(products , relabeling=relabeling, in_place=False)

        rxn = AnnotatedReaction.from_rdmols(reactant_templates=[reactants], product_templates=[products])
        rxn.Initialize()
        rxn.rxnname = self.rxn_name
        
        num_warnings, num_errors = rxn.Validate()
        if num_errors != 0:
            raise ValueError('Issues with reaction definition')

        return rxn


'''Decorators for modifying functions'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Callable, Concatenate, Iterable, Iterator, Optional, ParamSpec, TypeVar, Union

T = TypeVar('T')
Params = ParamSpec('Params')

from inspect import signature, Parameter
from functools import wraps, partial

from copy import deepcopy
from pathlib import Path

from .meta import extend_to_methods
from . import signatures
from ..fileutils.pathutils import aspath, asstrpath


@extend_to_methods
def optional_in_place(funct : Callable[[Concatenate[object, Params]], None]) -> Callable[[Concatenate[object, Params]], Optional[object]]:
    '''Decorator function for allowing in-place (writeable) functions which modify object attributes
    to be not performed in-place (i.e. read-only), specified by a boolean flag'''
    # TODO : add assertion that the wrapped function has at least one arg AND that the first arg is of the desired (limited) type
    old_sig = signature(funct)
    
    @wraps(funct) # for preserving docstring and type annotations / signatures
    def in_place_wrapper(obj : object, *args : Params.args, in_place : bool=False, **kwargs : Params.kwargs) -> Optional[object]: # read-only by default
        '''If not in-place, create a clone on which the method is executed''' # NOTE : old_sig.bind screws up arg passing
        if in_place:
            funct(obj, *args, **kwargs) # default call to writeable method - implicitly returns None
        else:
            copy_obj = deepcopy(obj) # clone object to avoid modifying original
            funct(copy_obj, *args, **kwargs) 
            return copy_obj # return the new object
    
    # ADD IN-PLACE PARAMETER TO FUNCTION SIGNATURE
    new_sig = signatures.insert_parameter_at_index(
        old_sig,
        new_param=Parameter(
            name='in_place',
            default=False,
            annotation=bool,
            kind=Parameter.KEYWORD_ONLY
        ),
        index=signatures.get_index_after_positionals(old_sig)
    )

    # ANNOTATE MODIFED OBJECT RETURN TYPE AS Optional[<type>]
    mod_type = tuple(old_sig.parameters.values())[0].annotation # get annotation of the object being modified
    in_place_wrapper.__signature__ = new_sig.replace(return_annotation=Optional[mod_type]) # replace signature with appropriate modifications

    return in_place_wrapper

# TODO : implement support for extend_to_methods (current mechanism is broken by additional deocrator parameters)
def flexible_listlike_input(funct : Callable[[Iterator], T]=None, CastType : type[Iterator]=list, valid_member_types : Union[type, tuple[type]]=object) -> Callable[[Iterable], T]:
    '''Wrapper which allows a function which expects a single list-initializable, Container-like object to accept any Iterable (or even star-unpacked arguments)'''
    if not issubclass(CastType, Iterator):
        raise TypeError(f'Cannot wrap listlike input with non-listlike type "{CastType.__name__}"')

    @wraps(funct)
    def wrapper(*args) -> T: # wrapper which accepts an arbitrary number of non-keyword argument
        if (len(args) == 1) and isinstance(args[0], Iterable):
            args = args[0]

        inputs = []
        for member in args:
            if isinstance(member, valid_member_types): # works because isinstance() accepts either a single type or a tuple of types
                inputs.append(member)
            else:
                raise TypeError(f'Item {member!r} of type {type(member).__name__} is not an instance of any of the following valid wrapped types: {valid_member_types}')
        inputs = CastType(inputs) # convert to the expected cast type (this is where the requirement of listlike cast types comes into play)

        return funct(inputs) # TODO: modify input type signature of wrapper function

    if funct is None:
        return partial(flexible_listlike_input, CastType=CastType, valid_member_types=valid_member_types)
    return wrapper

@extend_to_methods
def allow_string_paths(funct : Callable[[Concatenate[Path, Params]], T]) -> Callable[[Concatenate[Union[Path, str], Params]], T]:
    '''Modifies a function which expects a Path as its first argument to also accept string-paths'''
    # TODO : add assertion that the wrapped function has at least one arg AND that the first arg is of the desired (limited) type
    old_sig = signature(funct) # lookup old type signature

    @wraps(funct) # for preserving docstring and type annotations / signatures
    def str_path_wrapper(flex_path : Union[str, Path], *args : Params.args, **kwargs : Params.kwargs) -> T:
        '''First converts stringy paths into normal Paths, then executes the original function'''
        return funct(aspath(flex_path), *args, **kwargs)

    # MODIFY SIGNATURE OF PATH-LIKE FIRST ARGUMENT TO MATCH NEW TYPE FLEXIBILITY
    str_path_wrapper.__signature__ = signatures.modify_param_annotation_by_index(
        old_sig,
        index=0, # modify signature of first argument to reflect new type flexibility
        new_type=Union[Path, str]
    )

    return str_path_wrapper

@extend_to_methods
def allow_pathlib_paths(funct : Callable[[Concatenate[str, Params]], T]) -> Callable[[Concatenate[Union[Path, str], Params]], T]:
    '''Modifies a function which expects a string path as its first argument to also accept canonical pathlib Paths'''
    # TODO : add assertion that the wrapped function has at least one arg AND that the first arg is of the desired (limited) type
    old_sig = signature(funct) # lookup old type signature

    @wraps(funct) # for preserving docstring and type annotations / signatures
    def str_path_wrapper(flex_path : Union[str, Path], *args : Params.args, **kwargs : Params.kwargs) -> T:
        '''First converts normal Paths into stringy paths, then executes the original function'''
        return funct(asstrpath(flex_path), *args, **kwargs)

    # MODIFY SIGNATURE OF PATH-LIKE FIRST ARGUMENT TO MATCH NEW TYPE FLEXIBILITY
    str_path_wrapper.__signature__ = signatures.modify_param_annotation_by_index(
        old_sig,
        index=0, # modify signature of first argument to reflect new type flexibility
        new_type=Union[Path, str]
    )

    return str_path_wrapper


'''Tools for making existing classes easily readable/writable to JSON'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from typing import Any, Callable, ClassVar, Optional, Type, TypeVar, Union
C = TypeVar('C') # generic type for classes

from dataclasses import dataclass, is_dataclass
from functools import update_wrapper, wraps
from inspect import signature, isclass

import json
from pathlib import Path

from ...decorators.functional import allow_string_paths
from .serialize import TypeSerializer, MultiTypeSerializer


# TYPEHINTING AND SERIALIZATION FOR JSONIFIABLE CLASSES 
class JSONifiable: # documentation class, makes type-checking for jsonifiable-modified classes easier
    '''For type-hinting classes which are jsonifiable'''
    pass

def dataclass_serializer_factory(cls : Type[C]) -> TypeSerializer:
    '''For generating a custom TypeSerializer for a JSONifiable dataclass'''
    assert(is_dataclass(cls)) # can enforce modification only to dataclasses (makes behavior a little more natural)

    class DataclassSerializer(TypeSerializer, python_type=cls):
        f'''JSON encoder and decoder for the {cls.__name__} dataclass'''
        @staticmethod
        def encode(python_obj : Path) -> dict[str, Any]:
            '''Extract dictionary of attributes (may need other external converters to be fully serialized)'''
            return python_obj.__dict__

        @staticmethod
        def decode(json_obj : dict[str, Any]) -> C:
            '''Load registered JSONifiable class from a (presumed to be a dictionary during __values__ parse)'''
            json_obj = {
                attr : value
                    for attr, value in json_obj.items()
                        if cls.__dataclass_fields__[attr].init # only pass fields which are allowed to be initialized with values
            }
            return cls(**json_obj)
        
    # dynamically update signatures for readability
    non_generic_name = f'{cls.__name__}Serializer' # dynamically set the name of the serializer to match with the wrapped class
    setattr(DataclassSerializer, '__name__', non_generic_name)
    setattr(DataclassSerializer, '__qualname__', non_generic_name)
    setattr(DataclassSerializer, '__module__', cls.__module__)
        
    return DataclassSerializer


# DECORATOR METHOD FOR MAKING CLASSES JSON-SERIALIZABLE
def make_jsonifiable(cls : Optional[C]=None, type_serializer : Optional[Union[TypeSerializer, MultiTypeSerializer]]=None) -> C:
    '''
    Modify a dataclass to make its attributes writeable-to and readable-from JSON files
    Can optionally specify additional TypeSerializers to support objects with attributes whose types are, by default, not JSON-serializable
    '''
    def jsonifiable_factory(cls : C) -> C:
        '''Factory method, defines new class which inherits from original class and adds serialization methods'''
        assert(is_dataclass(cls)) # can enforce modification only to dataclasses (makes behavior a little more natural)

        multi_serializer = MultiTypeSerializer()
        if type_serializer is not None:
            multi_serializer.add_type_serializer(type_serializer)

        # check if any of the init fields of the dataclass are also JSONifiable, register respective serializers if they are
        for init_param in signature(cls).parameters.values(): # TODO: find a way to have this recognize serializable classes in default containers (e.g. list[JSONifiable])
            if isclass(init_param.annotation) and issubclass(init_param.annotation, JSONifiable): # check if the init field is itself a JSONifiable class
                multi_serializer.add_type_serializer(init_param.annotation.serializer)

        # generate serializable wrapper class
        @wraps(cls, updated=()) # copy over docstring, module, etc; set updated to empty so as to not attempt __dict__updates (classes don;t have these)
        @dataclass
        class WrappedClass(cls, JSONifiable):
            '''Class which inherits from modified class and adds JSON serialization capability'''
            serializer : ClassVar[MultiTypeSerializer] = multi_serializer

            @allow_string_paths
            def to_file(self, save_path : Path) -> None:
                '''Store parameters in a JSON file on disc'''
                assert(save_path.suffix == '.json')
                with save_path.open('w') as dumpfile:
                    json.dump(self, dumpfile, default=self.serializer.encoder_default, indent=4)

            @classmethod
            @allow_string_paths
            def from_file(cls, load_path : Path) -> cls.__class__:
                assert(load_path.suffix == '.json')
                with load_path.open('r') as loadfile:
                    return json.load(loadfile, object_hook=cls.serializer.decoder_hook)

        # !CRITICAL! that the custom serializer be registered for WrappedClass and NOT cls; otherwise, decoded instances will have different type to the parent class
        CustomSerializer = dataclass_serializer_factory(WrappedClass)
        multi_serializer.add_type_serializer(CustomSerializer)
        
        return WrappedClass

    if cls is None:
        return jsonifiable_factory
    return jsonifiable_factory(cls)


'''Strategies for redistribution excess partial charge among residues'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass
from abc import ABC, abstractmethod
from rdkit.Chem import Mol

from .rctypes import ChargeMap, ChargesByResidue


# ABSTRACT INTERFACE FOR DISTRIBUTING EXCESS RESIDUE CHARGES
@dataclass
class ChargeRedistributionStrategy(ABC):
    '''Interface for defining how any excess charge should be distributed within residues to ensure a given overall net charge'''
    desired_net_charge : float = 0.0 # by default, make neutral

    @abstractmethod # !NOTE! : this must be implemented in child classes
    def _determine_charge_offsets(self, base_charges : ChargeMap, fragment : Mol, net_charge_diff : float) -> ChargeMap:
        '''Provided a set of base charges, a structural molecule fragment, and a desired net charge,
        determine what charges offsets need to be applied where in order to achieve the desired net charge'''
        raise NotImplementedError

    def redistributed_charges(self, base_charges : ChargeMap, fragment : Mol) -> ChargeMap:
        '''Take a map of base charges and a structural fragment for a residue and a desired net charge (typically neutral, i.e. 0)
        and return a new charge map with the excess/deficit charge distributed in such a way as to make the residue have the desired net charge'''
        net_charge_diff = self.desired_net_charge - sum(chg for chg in base_charges.values())
        charge_offsets = self._determine_charge_offsets(base_charges, fragment, net_charge_diff)

        new_charges = {
            sub_id : charge + charge_offsets[sub_id]
                for sub_id, charge in base_charges.items()
        }

        # assert(sum(chg for chg in new_charges.values()) == desired_net_charge) # double check charges were correctly redistributed - TODO : find more reliable way to check this than floating-point comparison
        return new_charges

## CONCRETE IMPLEMENTATIONS OF REDISTRIBUTION STRATEGIES
class UniformDistributionStrategy(ChargeRedistributionStrategy):
    '''Simplest possible strategy, distribute any excess charge in a residue according to a uniform distribution (spread evenly)'''
    def _determine_charge_offsets(self, base_charges : ChargesByResidue, fragment : Mol, net_charge_diff : float) -> ChargesByResidue:
        return {sub_id : net_charge_diff / len(base_charges) for sub_id in base_charges}



'''
Enhanced conversions to and from mbuild Compound objects which
preserve more molecular information than the utilities provided by 
'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from ...genutils.importutils.dependencies import modules_installed, MissingPrerequisitePackage

if not modules_installed('openbabel'):
    raise MissingPrerequisitePackage(
        importing_package_name=__spec__.name,
        use_case='Translation between chemical representations of polymers',
        install_link='https://libraries.io/conda/openbabel',
        dependency_name='openbabel',
        dependency_name_formal='the OpenBabel chemical toolbox',
    )

from typing import Optional
from pathlib import Path

import warnings
with warnings.catch_warnings(record=True): # suppress numerous and irritating mbuild deprecation warnings
    warnings.filterwarnings('ignore',  category=DeprecationWarning)
    from mbuild import Compound
    from mbuild.conversion import from_rdkit
    
from rdkit import Chem

from ...genutils.decorators.functional import allow_string_paths, allow_pathlib_paths
from ..monomers.specification import SANITIZE_AS_KEKULE
from ...molfiles.pdb import SerialAtomLabeller
from ...rdutils.bonding.portlib import get_linker_ids
from ...rdutils.bonding.substitution import saturate_ports, hydrogenate_rdmol_ports
from ...mdtools.openmmtools.serialization import serialize_openmm_pdb


# Conversion from other formats to Compound
def mbmol_from_mono_rdmol(rdmol : Chem.Mol, resname : Optional[str]=None) -> tuple[Compound, list[int]]:
    '''
    Accepts a monomer-spec-compliant SMARTS string and returns an mbuild Compound and a list of the indices of atom ports
    If "resname" is provided, will assign that name to the mBuild Compound returned
    '''
    linker_ids = [i for i in get_linker_ids(rdmol)] # record indices of ports - MUST unpack generator for mbuild compatibility
    
    # create port-free version of molecule which RDKit can embed without errors
    prot_mol = hydrogenate_rdmol_ports(rdmol, in_place=False)
    # prot_mol = saturate_ports(rdmol) # TOSELF : custom, port-based saturation methods are not yet ready for deployment - yield issues in RDKit representation under-the-hood 
    Chem.SanitizeMol(prot_mol, sanitizeOps=SANITIZE_AS_KEKULE) # ensure Mol is valid (avoids implicitValence issues)
    
    mb_compound = from_rdkit(prot_mol) # native from_rdkit() method actually appears to preserve atom ordering
    if resname is not None:
        mb_compound.name = resname

    return mb_compound, linker_ids
   
# Conversion from Compound to other formats
_DEFAULT_RESNAME_MAP : dict[str, str] = { # module-wide config for default PDB residue name replacements for polymers
    'RES' : 'Pol',
}

def mbmol_to_rdmol( # TODO: deduplify PDB atom name and residue numbering code against serialize_openmm_pdb()
        mbmol : Compound,
        atom_labeller : Optional[SerialAtomLabeller]=SerialAtomLabeller(),
        resname_map : Optional[dict[str, str]]=None
    ) -> Chem.Mol:
    '''Convert an mBuild Compound into an RDKit Mol, with correct atom coordinates and PDB residue info'''
    if resname_map is None:
        resname_map = _DEFAULT_RESNAME_MAP
    
    rdmol = mbmol.to_rdkit()
    conformer = Chem.Conformer()
    conformer.Set3D(True)

    atom_id : int = 0
    for resnum, mb_monomer in enumerate(mbmol.children, start=1):
        resname = resname_map.get(mb_monomer.name, mb_monomer.name[:3]) # if no remapping is found, just take first 3 chars
        # NOTE: the order of monomers and atoms within those monomers were added in the same order as iterated over here...
        #... so the atom indices **SHOULD** be in the correct order (hate that this even might be uncertain)
        for mbatom in mb_monomer.particles(): 
            conformer.SetAtomPosition(atom_id, 10*mbatom.pos.astype(float)) # conveert from nm to angstrom

            # set PDB residue info if monomer hierarchy is present
            if mbatom != mb_monomer: # for Compounds with a flat hierarchy, the children and particles of children will coincide
                pdb_info = Chem.AtomPDBResidueInfo(
                    atomName=4*' ' if not atom_labeller else atom_labeller.get_atom_label(mbatom.element.symbol), 
                    residueName=resname,
                    residueNumber=resnum,
                    chainId='1',
                    isHeteroAtom=True,
                )
                rdmol.GetAtomWithIdx(atom_id).SetPDBResidueInfo(pdb_info)
            atom_id += 1 # TODO: this is an awful waay of keeping track of atom indices, see if there's a more secure way to do this
    conf_id = rdmol.AddConformer(conformer) # NOTE: recording this to self-document return values (this is intentionally not used)
    
    return rdmol

# Serialization of Compounds to files
@allow_pathlib_paths
def mbmol_to_rdkit_pdb(
        pdb_path : str,
        mbmol : Compound, 
        atom_labeller : Optional[SerialAtomLabeller]=SerialAtomLabeller(),
        resname_map : Optional[dict[str, str]]=None,
    ) -> None:
    '''Save an MBuild Compound into an RDKit-formatted PDB file'''
    Chem.MolToPDBFile(
        mbmol_to_rdmol(mbmol, atom_labeller=atom_labeller, resname_map=resname_map),
        pdb_path,
    )
    
@allow_string_paths
def mbmol_to_openmm_pdb(
        pdb_path : Path,
        mbmol : Compound, 
        atom_labeller : Optional[SerialAtomLabeller]=SerialAtomLabeller(),
        resname_map : Optional[dict[str, str]]=None,
    ) -> None:
    '''Save an MBuild Compound into an OpenMM-formatted PDB file'''
    if resname_map is None: # avoid mutable default
        resname_map = _DEFAULT_RESNAME_MAP 

     # NOTE: converting through MDTraj first before going to OpenMM preserves much
     # of the necessary chemical info that is discarded when converting through other formats
    traj = mbmol.to_trajectory(residues=[residue.name for residue in mbmol.children]) # extract names of repeat units
    omm_top, omm_pos = traj.top.to_openmm(), traj.openmm_positions(0) # extract OpenMM representations of trajectory

    serialize_openmm_pdb(
        pdb_path,
        topology=omm_top,
        positions=omm_pos,
        atom_labeller=atom_labeller,
        resname_map=resname_map,
    )


'''For recording, storing, and organizing parameters associated wtih a Simulation'''

__author__ = 'Timotej Bernat'
__email__ = 'timotej.bernat@colorado.edu'

from dataclasses import dataclass, field
from typing import Any

import numpy as np
from openmm.unit import Quantity

from .thermo import ThermoParameters
from .reporters import ReporterParameters
from ...genutils.fileutils.jsonio.jsonify import make_jsonifiable, dataclass_serializer_factory
from ...genutils.fileutils.jsonio.serialize import PathSerializer, QuantitySerializer, MultiTypeSerializer


# PARAMETER SET HANDLERS
@make_jsonifiable(type_serializer=QuantitySerializer)
@dataclass
class IntegratorParameters:
    '''For recording total, time step, recoridng frequency, and other integration time parameters'''
    time_step   : Quantity
    total_time  : Quantity
    num_samples : int

    @property
    def num_steps(self) -> int:
        '''Total number of steps in the simulation'''
        return round(self.total_time / self.time_step)
    
    @property
    def report_interval(self) -> int:
        '''Number of steps between successive samples'''
        return round(self.num_steps / self.num_samples)
    
    @property
    def report_duration(self) -> float:
        '''Length of time between successive samples'''
        return self.report_interval * self.time_step
    
    @property
    def time_points(self) -> np.ndarray[int]:
        '''An array of the time data points represented by the given sampling rate and sim time'''
        return (np.arange(0, self.num_steps, step=self.report_interval) + self.report_interval)* self.time_step # extra offset by recording frequency need to align indices (not 0-indexed)

# UNIFIED SIMULATION PARAMETER SETS
@make_jsonifiable
@dataclass
class SimulationParameters:
    '''Unified class for storing simulation parameters'''
    integ_params : IntegratorParameters
    thermo_params : ThermoParameters
    reporter_params : ReporterParameters
    
    def __post_init__(self) -> None:
        self.reporter_params.num_steps = self.integ_params.num_steps